import copy
import types
from typing import Sequence

from ..enums import DoneAction
from . import (
    DetectSilence,
    In,
    InFeedback,
    Line,
    Linen,
    LocalIn,
    LocalOut,
    Mix,
    Out,
    Parameter,
    RandID,
    ReplaceOut,
    SynthDefBuilder,
    UGenVector,
    XOut,
)


class SynthDefFactory:
    """
    A factory class for building SynthDefs with common signal flow structures.

    .. container:: example

        ::

            >>> from supriya.ugens.factories import SynthDefFactory
            >>> factory = SynthDefFactory()

        ::

            >>> def signal_block(builder, source, state):
            ...     iterations = state.get("iterations") or 2
            ...     for _ in range(iterations):
            ...         source = supriya.ugens.AllpassC.ar(
            ...             decay_time=supriya.ugens.ExpRand.ir(minimum=0.01, maximum=0.1),
            ...             delay_time=supriya.ugens.ExpRand.ir(minimum=0.01, maximum=0.1),
            ...             source=source,
            ...             maximum_delay_time=0.1,
            ...         )
            ...     return source

        ::

            >>> factory = factory.with_input()
            >>> factory = factory.with_output()
            >>> factory = factory.with_signal_block(signal_block)
            >>> synthdef = factory.build()
            >>> supriya.graph(synthdef)  # doctest: +SKIP

        ::

            >>> print(synthdef)
            synthdef:
                name: ...
                ugens:
                -   Control.ir:
                        out: 0.0
                -   In.ar:
                        bus: Control.ir[0:out]
                -   ExpRand.ir/0:
                        minimum: 0.01
                        maximum: 0.1
                -   ExpRand.ir/1:
                        minimum: 0.01
                        maximum: 0.1
                -   AllpassC.ar/0:
                        source: In.ar[0]
                        maximum_delay_time: 0.1
                        delay_time: ExpRand.ir/1[0]
                        decay_time: ExpRand.ir/0[0]
                -   ExpRand.ir/2:
                        minimum: 0.01
                        maximum: 0.1
                -   ExpRand.ir/3:
                        minimum: 0.01
                        maximum: 0.1
                -   AllpassC.ar/1:
                        source: AllpassC.ar/0[0]
                        maximum_delay_time: 0.1
                        delay_time: ExpRand.ir/3[0]
                        decay_time: ExpRand.ir/2[0]
                -   Out.ar:
                        bus: Control.ir[0:out]
                        source[0]: AllpassC.ar/1[0]

    .. container:: example

        ::

            >>> synthdef = factory.build(iterations=4)
            >>> supriya.graph(synthdef)  # doctest: +SKIP

        ::

            >>> print(synthdef)
            synthdef:
                name: ...
                ugens:
                -   Control.ir:
                        out: 0.0
                -   In.ar:
                        bus: Control.ir[0:out]
                -   ExpRand.ir/0:
                        minimum: 0.01
                        maximum: 0.1
                -   ExpRand.ir/1:
                        minimum: 0.01
                        maximum: 0.1
                -   AllpassC.ar/0:
                        source: In.ar[0]
                        maximum_delay_time: 0.1
                        delay_time: ExpRand.ir/1[0]
                        decay_time: ExpRand.ir/0[0]
                -   ExpRand.ir/2:
                        minimum: 0.01
                        maximum: 0.1
                -   ExpRand.ir/3:
                        minimum: 0.01
                        maximum: 0.1
                -   AllpassC.ar/1:
                        source: AllpassC.ar/0[0]
                        maximum_delay_time: 0.1
                        delay_time: ExpRand.ir/3[0]
                        decay_time: ExpRand.ir/2[0]
                -   ExpRand.ir/4:
                        minimum: 0.01
                        maximum: 0.1
                -   ExpRand.ir/5:
                        minimum: 0.01
                        maximum: 0.1
                -   AllpassC.ar/2:
                        source: AllpassC.ar/1[0]
                        maximum_delay_time: 0.1
                        delay_time: ExpRand.ir/5[0]
                        decay_time: ExpRand.ir/4[0]
                -   ExpRand.ir/6:
                        minimum: 0.01
                        maximum: 0.1
                -   ExpRand.ir/7:
                        minimum: 0.01
                        maximum: 0.1
                -   AllpassC.ar/3:
                        source: AllpassC.ar/2[0]
                        maximum_delay_time: 0.1
                        delay_time: ExpRand.ir/7[0]
                        decay_time: ExpRand.ir/6[0]
                -   Out.ar:
                        bus: Control.ir[0:out]
                        source[0]: AllpassC.ar/3[0]

    .. container:: example

        ::

            >>> synthdef = factory.build(channel_count=2)
            >>> supriya.graph(synthdef)  # doctest: +SKIP

        ::

            >>> print(synthdef)
            synthdef:
                name: ...
                ugens:
                -   Control.ir:
                        out: 0.0
                -   In.ar:
                        bus: Control.ir[0:out]
                -   ExpRand.ir/0:
                        minimum: 0.01
                        maximum: 0.1
                -   ExpRand.ir/1:
                        minimum: 0.01
                        maximum: 0.1
                -   AllpassC.ar/0:
                        source: In.ar[0]
                        maximum_delay_time: 0.1
                        delay_time: ExpRand.ir/1[0]
                        decay_time: ExpRand.ir/0[0]
                -   AllpassC.ar/1:
                        source: In.ar[1]
                        maximum_delay_time: 0.1
                        delay_time: ExpRand.ir/1[0]
                        decay_time: ExpRand.ir/0[0]
                -   ExpRand.ir/2:
                        minimum: 0.01
                        maximum: 0.1
                -   ExpRand.ir/3:
                        minimum: 0.01
                        maximum: 0.1
                -   AllpassC.ar/2:
                        source: AllpassC.ar/0[0]
                        maximum_delay_time: 0.1
                        delay_time: ExpRand.ir/3[0]
                        decay_time: ExpRand.ir/2[0]
                -   AllpassC.ar/3:
                        source: AllpassC.ar/1[0]
                        maximum_delay_time: 0.1
                        delay_time: ExpRand.ir/3[0]
                        decay_time: ExpRand.ir/2[0]
                -   Out.ar:
                        bus: Control.ir[0:out]
                        source[0]: AllpassC.ar/2[0]
                        source[1]: AllpassC.ar/3[0]
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_channel_count",
        "_feedback_loop",
        "_gate",
        "_initial_state",
        "_input",
        "_output",
        "_parameter_blocks",
        "_parameters",
        "_rand_id",
        "_signal_blocks",
        "_silence_detection",
    )

    ### INITIALIZER ###

    def __init__(self, channel_count=1, **kwargs):
        channel_count = int(channel_count)
        assert channel_count > 0
        self._channel_count = channel_count
        self._feedback_loop = None
        self._gate = {}
        self._initial_state = {}
        self._input = {}
        self._output = {}
        self._parameter_blocks = []
        self._parameters = sorted(tuple(kwargs.items()))
        self._rand_id = None
        self._signal_blocks = []
        self._silence_detection = None

    ### PRIVATE METHODS ###

    def _setup_parameters_and_state(self, builder, state, kwargs):
        state["channel_count"] = self._channel_count
        state.update(kwargs)
        for parameter_block in self._parameter_blocks:
            parameter_block(builder, state)
        if self._rand_id is not None:
            builder.add_parameter(name="rand_id", value=self._rand_id, rate="SCALAR")
            RandID.ir(rand_id=builder["rand_id"])
        if self._gate:
            builder.add_parameter(name="gate", value=1, rate="CONTROL")
            state["gate"] = Linen.kr(
                attack_time=self._gate["attack_time"],
                done_action=DoneAction.FREE_SYNTH,
                gate=builder["gate"],
                release_time=self._gate["release_time"],
            )
        if self._output or self._input:
            builder.add_parameter(name="out", value=0, rate="SCALAR")
        if self._input.get("private"):
            builder.add_parameter(name="in_", value=0, rate="SCALAR")
        if self._output.get("windowed") or self._input.get("windowed"):
            builder.add_parameter(name="duration", value=1, rate="SCALAR")
            state["line"] = Line.kr(
                done_action=DoneAction.FREE_SYNTH, duration=builder["duration"]
            )
            state["window"] = state["line"].hanning_window()
        if not self._output.get("windowed") and self._output.get("crossfaded"):
            builder.add_parameter(name="mix", value=0, rate="CONTROL")
        if self._output.get("leveled"):
            builder.add_parameter(name="level", value=1, rate="CONTROL")
        for key, value in self._parameters:
            builder.add_parameter(name=key, value=value)

    def _build_input(self, builder, state):
        if not self._input:
            return
        parameter = builder["out"]
        if self._input.get("private"):
            parameter = builder["in_"]
        input_class = In
        if self._input.get("feedback"):
            input_class = InFeedback
        source = input_class.ar(bus=parameter, channel_count=state["channel_count"])
        if self._input.get("windowed"):
            source *= state["window"]
        if source is not None and not isinstance(source, Sequence):
            source = UGenVector(source)
        return source

    def _build_feedback_loop_input(self, builder, source, state):
        if self._feedback_loop:
            local_in = LocalIn.ar(channel_count=state["channel_count"])
            if source is None:
                source = local_in
            else:
                source += local_in
        if source is not None and not isinstance(source, Sequence):
            source = UGenVector(source)
        return source

    def _build_feedback_loop_output(self, builder, source, state):
        if not self._feedback_loop:
            return
        if isinstance(self._feedback_loop, types.FunctionType):
            source = self._feedback_loop(builder, source, state)
        LocalOut.ar(source=source)

    def _build_output(self, builder, source, state):
        if not self._output:
            return
        crossfaded = self._output.get("crossfaded")
        replacing = self._output.get("replacing")
        windowed = self._output.get("windowed")
        gate = state.get("gate")
        if self._output.get("leveled") and not crossfaded:
            source *= builder["level"]
        out_class = Out
        kwargs = dict(bus=builder["out"], source=source)
        if replacing:
            out_class = ReplaceOut
        if crossfaded:
            out_class = XOut
            if windowed:
                window = state["window"]
                if self._output.get("leveled"):
                    window *= builder["level"]
                kwargs["crossfade"] = window
            else:
                kwargs["crossfade"] = builder["mix"]
            if gate:
                kwargs["crossfade"] *= gate
        elif windowed:
            window = state["window"]
            kwargs["source"] *= window
            if gate:
                kwargs["source"] *= gate
        elif gate:
            kwargs["source"] *= gate
        out_class.ar(**kwargs)

    def _build_silence_detection(self, builder, source, state):
        if not self._silence_detection:
            return
        DetectSilence.kr(done_action=DoneAction.FREE_SYNTH, source=Mix.new(source))

    def _clone(self):
        clone = type(self)()
        for name in self.__slots__:
            value = getattr(self, name)
            setattr(clone, name, copy.copy(value))
        return clone

    ### PUBLIC METHODS ###

    def build(self, name=None, **kwargs):
        """
        Build the SynthDef.
        """
        builder = SynthDefBuilder()
        state = self._initial_state.copy()
        with builder:
            state.update(**kwargs)
            self._setup_parameters_and_state(builder, state, kwargs)
            source = self._build_input(builder, state)
            source = self._build_feedback_loop_input(builder, source, state)
            for signal_block in self._signal_blocks:
                source = signal_block(builder, source, state)
                if source is not None and not isinstance(source, Sequence):
                    source = UGenVector(source)
            self._build_output(builder, source, state)
            self._build_feedback_loop_output(builder, source, state)
            self._build_silence_detection(builder, source, state)
        return builder.build(name=name)

    def with_channel_count(self, channel_count):
        """
        Return a new factory configured with `channel_count`.

        .. container:: example

            ::

                >>> from supriya.ugens.factories import SynthDefFactory
                >>> factory = SynthDefFactory()

            ::

                >>> def signal_block(builder, source, state):
                ...     iterations = state.get("iterations") or 2
                ...     for _ in range(iterations):
                ...         source = supriya.ugens.AllpassC.ar(
                ...             decay_time=supriya.ugens.ExpRand.ir(minimum=0.01, maximum=0.1),
                ...             delay_time=supriya.ugens.ExpRand.ir(minimum=0.01, maximum=0.1),
                ...             source=source,
                ...             maximum_delay_time=0.1,
                ...         )
                ...     return source

            ::

                >>> factory = factory.with_input()
                >>> factory = factory.with_output()
                >>> factory = factory.with_signal_block(signal_block)

        .. container:: example

            Configure the factory with 4 channels:

            ::

                >>> factory = factory.with_channel_count(4)
                >>> synthdef = factory.build()
                >>> supriya.graph(synthdef)  # doctest: +SKIP

            ::

                >>> print(synthdef)
                synthdef:
                    name: ...
                    ugens:
                    -   Control.ir:
                            out: 0.0
                    -   In.ar:
                            bus: Control.ir[0:out]
                    -   ExpRand.ir/0:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/1:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/0:
                            source: In.ar[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/1[0]
                            decay_time: ExpRand.ir/0[0]
                    -   AllpassC.ar/1:
                            source: In.ar[1]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/1[0]
                            decay_time: ExpRand.ir/0[0]
                    -   AllpassC.ar/2:
                            source: In.ar[2]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/1[0]
                            decay_time: ExpRand.ir/0[0]
                    -   AllpassC.ar/3:
                            source: In.ar[3]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/1[0]
                            decay_time: ExpRand.ir/0[0]
                    -   ExpRand.ir/2:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/3:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/4:
                            source: AllpassC.ar/0[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/3[0]
                            decay_time: ExpRand.ir/2[0]
                    -   AllpassC.ar/5:
                            source: AllpassC.ar/1[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/3[0]
                            decay_time: ExpRand.ir/2[0]
                    -   AllpassC.ar/6:
                            source: AllpassC.ar/2[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/3[0]
                            decay_time: ExpRand.ir/2[0]
                    -   AllpassC.ar/7:
                            source: AllpassC.ar/3[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/3[0]
                            decay_time: ExpRand.ir/2[0]
                    -   Out.ar:
                            bus: Control.ir[0:out]
                            source[0]: AllpassC.ar/4[0]
                            source[1]: AllpassC.ar/5[0]
                            source[2]: AllpassC.ar/6[0]
                            source[3]: AllpassC.ar/7[0]

        .. container:: example

            Channel count can be overridden at build time:

            ::

                >>> synthdef = factory.build(channel_count=3)
                >>> supriya.graph(synthdef)  # doctest: +SKIP

            ::

                >>> print(synthdef)
                synthdef:
                    name: ...
                    ugens:
                    -   Control.ir:
                            out: 0.0
                    -   In.ar:
                            bus: Control.ir[0:out]
                    -   ExpRand.ir/0:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/1:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/0:
                            source: In.ar[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/1[0]
                            decay_time: ExpRand.ir/0[0]
                    -   AllpassC.ar/1:
                            source: In.ar[1]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/1[0]
                            decay_time: ExpRand.ir/0[0]
                    -   AllpassC.ar/2:
                            source: In.ar[2]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/1[0]
                            decay_time: ExpRand.ir/0[0]
                    -   ExpRand.ir/2:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/3:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/3:
                            source: AllpassC.ar/0[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/3[0]
                            decay_time: ExpRand.ir/2[0]
                    -   AllpassC.ar/4:
                            source: AllpassC.ar/1[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/3[0]
                            decay_time: ExpRand.ir/2[0]
                    -   AllpassC.ar/5:
                            source: AllpassC.ar/2[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/3[0]
                            decay_time: ExpRand.ir/2[0]
                    -   Out.ar:
                            bus: Control.ir[0:out]
                            source[0]: AllpassC.ar/3[0]
                            source[1]: AllpassC.ar/4[0]
                            source[2]: AllpassC.ar/5[0]
        """
        channel_count = int(channel_count)
        assert channel_count > 0
        clone = self._clone()
        clone._channel_count = channel_count
        return clone

    def with_feedback_loop(self, block_function=None):
        """
        Return a new factory configured with a feedback loop.

        Feedback block functions follow the same guidelines as other signal block
        functions.

        .. container:: example

            ::

                >>> from supriya.ugens.factories import SynthDefFactory
                >>> factory = SynthDefFactory()

            ::

                >>> def signal_block(builder, source, state):
                ...     iterations = state.get("iterations") or 2
                ...     for _ in range(iterations):
                ...         source = supriya.ugens.AllpassC.ar(
                ...             decay_time=supriya.ugens.ExpRand.ir(minimum=0.01, maximum=0.1),
                ...             delay_time=supriya.ugens.ExpRand.ir(minimum=0.01, maximum=0.1),
                ...             source=source,
                ...             maximum_delay_time=0.1,
                ...         )
                ...     return source

            ::

                >>> factory = factory.with_input()
                >>> factory = factory.with_output()
                >>> factory = factory.with_signal_block(signal_block)

        .. container:: example

            Configure the factory with a basic feedback loop:

            ::

                >>> factory = factory.with_feedback_loop()
                >>> synthdef = factory.build()
                >>> supriya.graph(synthdef)  # doctest: +SKIP

            ::

                >>> print(synthdef)
                synthdef:
                    name: ...
                    ugens:
                    -   Control.ir:
                            out: 0.0
                    -   In.ar:
                            bus: Control.ir[0:out]
                    -   LocalIn.ar:
                            default[0]: 0.0
                    -   BinaryOpUGen(ADDITION).ar:
                            left: In.ar[0]
                            right: LocalIn.ar[0]
                    -   ExpRand.ir/0:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/1:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/0:
                            source: BinaryOpUGen(ADDITION).ar[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/1[0]
                            decay_time: ExpRand.ir/0[0]
                    -   ExpRand.ir/2:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/3:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/1:
                            source: AllpassC.ar/0[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/3[0]
                            decay_time: ExpRand.ir/2[0]
                    -   Out.ar:
                            bus: Control.ir[0:out]
                            source[0]: AllpassC.ar/1[0]
                    -   LocalOut.ar:
                            source[0]: AllpassC.ar/1[0]

        .. container:: example

            Configure the factory with a modulated feedback loop via a signal block
            function:

            ::

                >>> def feedback_block(builder, source, state):
                ...     return source * supriya.ugens.SinOsc.kr(frequency=0.3)

            ::

                >>> factory = factory.with_feedback_loop(feedback_block)
                >>> synthdef = factory.build()
                >>> supriya.graph(synthdef)  # doctest: +SKIP

            ::

                >>> print(synthdef)
                synthdef:
                    name: ...
                    ugens:
                    -   Control.ir:
                            out: 0.0
                    -   In.ar:
                            bus: Control.ir[0:out]
                    -   LocalIn.ar:
                            default[0]: 0.0
                    -   BinaryOpUGen(ADDITION).ar:
                            left: In.ar[0]
                            right: LocalIn.ar[0]
                    -   ExpRand.ir/0:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/1:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/0:
                            source: BinaryOpUGen(ADDITION).ar[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/1[0]
                            decay_time: ExpRand.ir/0[0]
                    -   ExpRand.ir/2:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/3:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/1:
                            source: AllpassC.ar/0[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/3[0]
                            decay_time: ExpRand.ir/2[0]
                    -   Out.ar:
                            bus: Control.ir[0:out]
                            source[0]: AllpassC.ar/1[0]
                    -   SinOsc.kr:
                            frequency: 0.3
                            phase: 0.0
                    -   BinaryOpUGen(MULTIPLICATION).ar:
                            left: AllpassC.ar/1[0]
                            right: SinOsc.kr[0]
                    -   LocalOut.ar:
                            source[0]: BinaryOpUGen(MULTIPLICATION).ar[0]
        """
        clone = self._clone()
        if block_function:
            clone._feedback_loop = block_function
        else:
            clone._feedback_loop = True
        return clone

    def with_gate(self, attack_time=0.02, release_time=0.02):
        """
        Return a new factory configured with a gate.

        .. container:: example

            ::

                >>> from supriya.ugens.factories import SynthDefFactory
                >>> factory = SynthDefFactory()

            ::

                >>> def signal_block(builder, source, state):
                ...     iterations = state.get("iterations") or 2
                ...     for _ in range(iterations):
                ...         source = supriya.ugens.AllpassC.ar(
                ...             decay_time=supriya.ugens.ExpRand.ir(minimum=0.01, maximum=0.1),
                ...             delay_time=supriya.ugens.ExpRand.ir(minimum=0.01, maximum=0.1),
                ...             source=source,
                ...             maximum_delay_time=0.1,
                ...         )
                ...     return source

            ::

                >>> factory = factory.with_input()
                >>> factory = factory.with_output()
                >>> factory = factory.with_signal_block(signal_block)

        .. container:: example

            Configure the factory with a gate envelope and corresponding gate parameter:

            ::

                >>> factory = factory.with_gate()
                >>> synthdef = factory.build()
                >>> supriya.graph(synthdef)  # doctest: +SKIP

            ::

                >>> print(synthdef)
                synthdef:
                    name: ...
                    ugens:
                    -   Control.kr:
                            gate: 1.0
                    -   Linen.kr:
                            gate: Control.kr[0:gate]
                            attack_time: 0.02
                            sustain_level: 1.0
                            release_time: 0.02
                            done_action: 2.0
                    -   Control.ir:
                            out: 0.0
                    -   In.ar:
                            bus: Control.ir[0:out]
                    -   ExpRand.ir/0:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/1:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/0:
                            source: In.ar[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/1[0]
                            decay_time: ExpRand.ir/0[0]
                    -   ExpRand.ir/2:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/3:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/1:
                            source: AllpassC.ar/0[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/3[0]
                            decay_time: ExpRand.ir/2[0]
                    -   BinaryOpUGen(MULTIPLICATION).ar:
                            left: AllpassC.ar/1[0]
                            right: Linen.kr[0]
                    -   Out.ar:
                            bus: Control.ir[0:out]
                            source[0]: BinaryOpUGen(MULTIPLICATION).ar[0]
        """
        clone = self._clone()
        clone._gate.update(
            attack_time=float(attack_time), release_time=float(release_time)
        )
        return clone

    def with_initial_state(self, **state):
        """
        Return a new factory configured with an inital state comprised of key/value pairs.

        .. container:: example

            ::

                >>> from supriya.ugens.factories import SynthDefFactory
                >>> factory = SynthDefFactory()

            ::

                >>> def signal_block(builder, source, state):
                ...     iterations = state.get("iterations") or 2
                ...     for _ in range(iterations):
                ...         source = supriya.ugens.AllpassC.ar(
                ...             decay_time=supriya.ugens.ExpRand.ir(minimum=0.01, maximum=0.1),
                ...             delay_time=supriya.ugens.ExpRand.ir(minimum=0.01, maximum=0.1),
                ...             source=source,
                ...             maximum_delay_time=0.1,
                ...         )
                ...     return source

            ::

                >>> factory = factory.with_input()
                >>> factory = factory.with_output()
                >>> factory = factory.with_signal_block(signal_block)

        .. container:: example

            Configure the factory with an initial state consisting of a single key/value
            pair which can be accessed in the previously configured signal block
            function:

            ::

                >>> factory = factory.with_initial_state(iterations=4)
                >>> synthdef = factory.build()
                >>> supriya.graph(synthdef)  # doctest: +SKIP

            ::

                >>> print(synthdef)
                synthdef:
                    name: ...
                    ugens:
                    -   Control.ir:
                            out: 0.0
                    -   In.ar:
                            bus: Control.ir[0:out]
                    -   ExpRand.ir/0:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/1:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/0:
                            source: In.ar[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/1[0]
                            decay_time: ExpRand.ir/0[0]
                    -   ExpRand.ir/2:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/3:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/1:
                            source: AllpassC.ar/0[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/3[0]
                            decay_time: ExpRand.ir/2[0]
                    -   ExpRand.ir/4:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/5:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/2:
                            source: AllpassC.ar/1[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/5[0]
                            decay_time: ExpRand.ir/4[0]
                    -   ExpRand.ir/6:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/7:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/3:
                            source: AllpassC.ar/2[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/7[0]
                            decay_time: ExpRand.ir/6[0]
                    -   Out.ar:
                            bus: Control.ir[0:out]
                            source[0]: AllpassC.ar/3[0]
        """
        clone = self._clone()
        clone._initial_state.update(**state)
        return clone

    def with_input(self, feedback=False, private=False, windowed=False):
        """
        Return a new factory configured with a bus input.

        .. container:: example

            ::

                >>> from supriya.ugens.factories import SynthDefFactory
                >>> factory = SynthDefFactory()

            ::

                >>> def signal_block(builder, source, state):
                ...     iterations = state.get("iterations") or 2
                ...     for _ in range(iterations):
                ...         source = supriya.ugens.AllpassC.ar(
                ...             decay_time=supriya.ugens.ExpRand.ir(minimum=0.01, maximum=0.1),
                ...             delay_time=supriya.ugens.ExpRand.ir(minimum=0.01, maximum=0.1),
                ...             source=source,
                ...             maximum_delay_time=0.1,
                ...         )
                ...     return source

            ::

                >>> factory = factory.with_signal_block(signal_block)
                >>> factory = factory.with_output()

        .. container:: example

            Configure the factory with a basic bus input:

            ::

                >>> factory = factory.with_input()
                >>> synthdef = factory.build()
                >>> supriya.graph(synthdef)  # doctest: +SKIP

            ::

                >>> print(synthdef)
                synthdef:
                    name: ...
                    ugens:
                    -   Control.ir:
                            out: 0.0
                    -   In.ar:
                            bus: Control.ir[0:out]
                    -   ExpRand.ir/0:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/1:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/0:
                            source: In.ar[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/1[0]
                            decay_time: ExpRand.ir/0[0]
                    -   ExpRand.ir/2:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/3:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/1:
                            source: AllpassC.ar/0[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/3[0]
                            decay_time: ExpRand.ir/2[0]
                    -   Out.ar:
                            bus: Control.ir[0:out]
                            source[0]: AllpassC.ar/1[0]

        .. container:: example

            Configure the factory with a private bus input:

            ::

                >>> factory = factory.with_input(private=True)
                >>> synthdef = factory.build()
                >>> supriya.graph(synthdef)  # doctest: +SKIP

            ::

                >>> print(synthdef)
                synthdef:
                    name: ...
                    ugens:
                    -   Control.ir:
                            in_: 0.0
                            out: 0.0
                    -   In.ar:
                            bus: Control.ir[0:in_]
                    -   ExpRand.ir/0:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/1:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/0:
                            source: In.ar[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/1[0]
                            decay_time: ExpRand.ir/0[0]
                    -   ExpRand.ir/2:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/3:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/1:
                            source: AllpassC.ar/0[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/3[0]
                            decay_time: ExpRand.ir/2[0]
                    -   Out.ar:
                            bus: Control.ir[1:out]
                            source[0]: AllpassC.ar/1[0]

        .. container:: example

            Configure the factory with a windowed bus input:

            ::

                >>> factory = factory.with_input(windowed=True)
                >>> synthdef = factory.build()
                >>> supriya.graph(synthdef)  # doctest: +SKIP

            ::

                >>> print(synthdef)
                synthdef:
                    name: ...
                    ugens:
                    -   Control.ir:
                            duration: 1.0
                            out: 0.0
                    -   Line.kr:
                            start: 0.0
                            stop: 1.0
                            duration: Control.ir[0:duration]
                            done_action: 2.0
                    -   UnaryOpUGen(HANNING_WINDOW).kr:
                            source: Line.kr[0]
                    -   In.ar:
                            bus: Control.ir[1:out]
                    -   BinaryOpUGen(MULTIPLICATION).ar:
                            left: In.ar[0]
                            right: UnaryOpUGen(HANNING_WINDOW).kr[0]
                    -   ExpRand.ir/0:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/1:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/0:
                            source: BinaryOpUGen(MULTIPLICATION).ar[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/1[0]
                            decay_time: ExpRand.ir/0[0]
                    -   ExpRand.ir/2:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/3:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/1:
                            source: AllpassC.ar/0[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/3[0]
                            decay_time: ExpRand.ir/2[0]
                    -   Out.ar:
                            bus: Control.ir[1:out]
                            source[0]: AllpassC.ar/1[0]

        .. container:: example

            A factory configured with both a windowed bus input and output will re-use
            the windowing signal:

            ::

                >>> factory = factory.with_input(windowed=True)
                >>> factory = factory.with_output(windowed=True)
                >>> synthdef = factory.build()
                >>> supriya.graph(synthdef)  # doctest: +SKIP

            ::

                >>> print(synthdef)
                synthdef:
                    name: ...
                    ugens:
                    -   Control.ir:
                            duration: 1.0
                            out: 0.0
                    -   Line.kr:
                            start: 0.0
                            stop: 1.0
                            duration: Control.ir[0:duration]
                            done_action: 2.0
                    -   UnaryOpUGen(HANNING_WINDOW).kr:
                            source: Line.kr[0]
                    -   In.ar:
                            bus: Control.ir[1:out]
                    -   BinaryOpUGen(MULTIPLICATION).ar/0:
                            left: In.ar[0]
                            right: UnaryOpUGen(HANNING_WINDOW).kr[0]
                    -   ExpRand.ir/0:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/1:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/0:
                            source: BinaryOpUGen(MULTIPLICATION).ar/0[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/1[0]
                            decay_time: ExpRand.ir/0[0]
                    -   ExpRand.ir/2:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/3:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/1:
                            source: AllpassC.ar/0[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/3[0]
                            decay_time: ExpRand.ir/2[0]
                    -   BinaryOpUGen(MULTIPLICATION).ar/1:
                            left: AllpassC.ar/1[0]
                            right: UnaryOpUGen(HANNING_WINDOW).kr[0]
                    -   Out.ar:
                            bus: Control.ir[1:out]
                            source[0]: BinaryOpUGen(MULTIPLICATION).ar/1[0]
        """
        clone = self._clone()
        clone._input.update(
            feedback=bool(feedback), private=bool(private), windowed=bool(windowed)
        )
        return clone

    def with_output(
        self, crossfaded=False, leveled=False, replacing=False, windowed=False
    ):
        """
        Return a new factory configured with a bus output.

        .. container:: example

            ::

                >>> from supriya.ugens.factories import SynthDefFactory
                >>> factory = SynthDefFactory()

            ::

                >>> def signal_block(builder, source, state):
                ...     iterations = state.get("iterations") or 2
                ...     for _ in range(iterations):
                ...         source = supriya.ugens.AllpassC.ar(
                ...             decay_time=supriya.ugens.ExpRand.ir(minimum=0.01, maximum=0.1),
                ...             delay_time=supriya.ugens.ExpRand.ir(minimum=0.01, maximum=0.1),
                ...             source=source,
                ...             maximum_delay_time=0.1,
                ...         )
                ...     return source

            ::

                >>> factory = factory.with_input()
                >>> factory = factory.with_signal_block(signal_block)

        .. container:: example

            Configure the factory with a basic bus output:

            ::

                >>> factory = factory.with_output()
                >>> synthdef = factory.build()
                >>> supriya.graph(synthdef)  # doctest: +SKIP

            ::

                >>> print(synthdef)
                synthdef:
                    name: ...
                    ugens:
                    -   Control.ir:
                            out: 0.0
                    -   In.ar:
                            bus: Control.ir[0:out]
                    -   ExpRand.ir/0:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/1:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/0:
                            source: In.ar[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/1[0]
                            decay_time: ExpRand.ir/0[0]
                    -   ExpRand.ir/2:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/3:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/1:
                            source: AllpassC.ar/0[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/3[0]
                            decay_time: ExpRand.ir/2[0]
                    -   Out.ar:
                            bus: Control.ir[0:out]
                            source[0]: AllpassC.ar/1[0]

        .. container:: example

            Configure the factory with a windowed bus output:

            ::

                >>> factory = factory.with_output(windowed=True)
                >>> synthdef = factory.build()
                >>> supriya.graph(synthdef)  # doctest: +SKIP

            ::

                >>> print(synthdef)
                synthdef:
                    name: ...
                    ugens:
                    -   Control.ir:
                            duration: 1.0
                            out: 0.0
                    -   Line.kr:
                            start: 0.0
                            stop: 1.0
                            duration: Control.ir[0:duration]
                            done_action: 2.0
                    -   UnaryOpUGen(HANNING_WINDOW).kr:
                            source: Line.kr[0]
                    -   In.ar:
                            bus: Control.ir[1:out]
                    -   ExpRand.ir/0:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/1:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/0:
                            source: In.ar[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/1[0]
                            decay_time: ExpRand.ir/0[0]
                    -   ExpRand.ir/2:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/3:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/1:
                            source: AllpassC.ar/0[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/3[0]
                            decay_time: ExpRand.ir/2[0]
                    -   BinaryOpUGen(MULTIPLICATION).ar:
                            left: AllpassC.ar/1[0]
                            right: UnaryOpUGen(HANNING_WINDOW).kr[0]
                    -   Out.ar:
                            bus: Control.ir[1:out]
                            source[0]: BinaryOpUGen(MULTIPLICATION).ar[0]

        .. container:: example

            Configure the factory with a mix-able bus output:

            ::

                >>> factory = factory.with_output(crossfaded=True)
                >>> synthdef = factory.build()
                >>> supriya.graph(synthdef)  # doctest: +SKIP

            ::

                >>> print(synthdef)
                synthdef:
                    name: ...
                    ugens:
                    -   Control.kr:
                            mix: 0.0
                    -   Control.ir:
                            out: 0.0
                    -   In.ar:
                            bus: Control.ir[0:out]
                    -   ExpRand.ir/0:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/1:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/0:
                            source: In.ar[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/1[0]
                            decay_time: ExpRand.ir/0[0]
                    -   ExpRand.ir/2:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/3:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/1:
                            source: AllpassC.ar/0[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/3[0]
                            decay_time: ExpRand.ir/2[0]
                    -   XOut.ar:
                            bus: Control.ir[0:out]
                            crossfade: Control.kr[0:mix]
                            source[0]: AllpassC.ar/1[0]

        .. container:: example

            Configure the factory with a basic bus output preceded by an amplitude level
            control:

            ::

                >>> factory = factory.with_output(leveled=True)
                >>> synthdef = factory.build()
                >>> supriya.graph(synthdef)  # doctest: +SKIP

            ::

                >>> print(synthdef)
                synthdef:
                    name: ...
                    ugens:
                    -   Control.kr:
                            level: 1.0
                    -   Control.ir:
                            out: 0.0
                    -   In.ar:
                            bus: Control.ir[0:out]
                    -   ExpRand.ir/0:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/1:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/0:
                            source: In.ar[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/1[0]
                            decay_time: ExpRand.ir/0[0]
                    -   ExpRand.ir/2:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/3:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/1:
                            source: AllpassC.ar/0[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/3[0]
                            decay_time: ExpRand.ir/2[0]
                    -   BinaryOpUGen(MULTIPLICATION).ar:
                            left: AllpassC.ar/1[0]
                            right: Control.kr[0:level]
                    -   Out.ar:
                            bus: Control.ir[0:out]
                            source[0]: BinaryOpUGen(MULTIPLICATION).ar[0]

        .. container:: example

            A factory configured with a crossfaded *and* windowed bus output will use
            the windowing signal to control the mix:

            ::

                >>> factory = factory.with_output(
                ...     crossfaded=True,
                ...     windowed=True,
                ... )
                >>> synthdef = factory.build()
                >>> supriya.graph(synthdef)  # doctest: +SKIP

            ::

                >>> print(synthdef)
                synthdef:
                    name: ...
                    ugens:
                    -   Control.ir:
                            duration: 1.0
                            out: 0.0
                    -   Line.kr:
                            start: 0.0
                            stop: 1.0
                            duration: Control.ir[0:duration]
                            done_action: 2.0
                    -   UnaryOpUGen(HANNING_WINDOW).kr:
                            source: Line.kr[0]
                    -   In.ar:
                            bus: Control.ir[1:out]
                    -   ExpRand.ir/0:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/1:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/0:
                            source: In.ar[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/1[0]
                            decay_time: ExpRand.ir/0[0]
                    -   ExpRand.ir/2:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/3:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/1:
                            source: AllpassC.ar/0[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/3[0]
                            decay_time: ExpRand.ir/2[0]
                    -   XOut.ar:
                            bus: Control.ir[1:out]
                            crossfade: UnaryOpUGen(HANNING_WINDOW).kr[0]
                            source[0]: AllpassC.ar/1[0]

        .. container:: example

            A level-control can be combined with the windowing and crossfading:

            ::

                >>> factory = factory.with_output(
                ...     crossfaded=True,
                ...     leveled=True,
                ...     windowed=True,
                ... )
                >>> synthdef = factory.build()
                >>> supriya.graph(synthdef)  # doctest: +SKIP

            ::

                >>> print(synthdef)
                synthdef:
                    name: ...
                    ugens:
                    -   Control.kr:
                            level: 1.0
                    -   Control.ir:
                            duration: 1.0
                            out: 0.0
                    -   Line.kr:
                            start: 0.0
                            stop: 1.0
                            duration: Control.ir[0:duration]
                            done_action: 2.0
                    -   UnaryOpUGen(HANNING_WINDOW).kr:
                            source: Line.kr[0]
                    -   BinaryOpUGen(MULTIPLICATION).kr:
                            left: UnaryOpUGen(HANNING_WINDOW).kr[0]
                            right: Control.kr[0:level]
                    -   In.ar:
                            bus: Control.ir[1:out]
                    -   ExpRand.ir/0:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/1:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/0:
                            source: In.ar[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/1[0]
                            decay_time: ExpRand.ir/0[0]
                    -   ExpRand.ir/2:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/3:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/1:
                            source: AllpassC.ar/0[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/3[0]
                            decay_time: ExpRand.ir/2[0]
                    -   XOut.ar:
                            bus: Control.ir[1:out]
                            crossfade: BinaryOpUGen(MULTIPLICATION).kr[0]
                            source[0]: AllpassC.ar/1[0]

        .. container:: example

            A factory configured with both a windowed bus input and output will re-use
            the windowing signal:

            ::

                >>> factory = factory.with_input(windowed=True)
                >>> synthdef = factory.build()
                >>> supriya.graph(synthdef)  # doctest: +SKIP

            ::

                >>> print(synthdef)
                synthdef:
                    name: ...
                    ugens:
                    -   Control.kr:
                            level: 1.0
                    -   Control.ir:
                            duration: 1.0
                            out: 0.0
                    -   Line.kr:
                            start: 0.0
                            stop: 1.0
                            duration: Control.ir[0:duration]
                            done_action: 2.0
                    -   UnaryOpUGen(HANNING_WINDOW).kr:
                            source: Line.kr[0]
                    -   BinaryOpUGen(MULTIPLICATION).kr:
                            left: UnaryOpUGen(HANNING_WINDOW).kr[0]
                            right: Control.kr[0:level]
                    -   In.ar:
                            bus: Control.ir[1:out]
                    -   BinaryOpUGen(MULTIPLICATION).ar:
                            left: In.ar[0]
                            right: UnaryOpUGen(HANNING_WINDOW).kr[0]
                    -   ExpRand.ir/0:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/1:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/0:
                            source: BinaryOpUGen(MULTIPLICATION).ar[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/1[0]
                            decay_time: ExpRand.ir/0[0]
                    -   ExpRand.ir/2:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/3:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/1:
                            source: AllpassC.ar/0[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/3[0]
                            decay_time: ExpRand.ir/2[0]
                    -   XOut.ar:
                            bus: Control.ir[1:out]
                            crossfade: BinaryOpUGen(MULTIPLICATION).kr[0]
                            source[0]: AllpassC.ar/1[0]
        """
        assert not (replacing and crossfaded)
        clone = self._clone()
        clone._output.update(
            crossfaded=bool(crossfaded),
            leveled=bool(leveled),
            replacing=bool(replacing),
            windowed=bool(windowed),
        )
        return clone

    def with_parameter(self, name, value, rate=None):
        parameter = Parameter(name=name, value=value, rate=rate)
        return self.with_parameters(**{name: parameter})

    def with_parameters(self, **kwargs):
        clone = self._clone()
        parameters = dict(self._parameters)
        for key, value in kwargs.items():
            if key in parameters and value is None:
                parameters.pop(key)
            else:
                parameters[key] = value
        clone._parameters = sorted(tuple(parameters.items()))
        return clone

    def with_parameter_block(self, block_function):
        """
        Return a new factory configured with a parameter block function.

        Use parameter block functions to build repetitive sets of SynthDef parameters,
        e.g. each set of band parameters for a multi-band compressor.

        Parameter block functions take two parameters:

        builder
            the SynthDef builder instance

        state
            a dictionary of arbitrary key/value pairs for parameterizing the signal and
            parameter block functions

        The return values of parameter block functions are ignored.

        .. container:: example

            A factory configured to build multi-band compressor SynthDefs, using
            frequency bands split at ``frequencies``:

            ::

                >>> from typing import Sequence
                >>> from supriya.ugens import CompanderD, LPF, Mix

            ::

                >>> def parameter_block(builder, state):
                ...     frequencies = state["frequencies"]
                ...     band_count = len(frequencies) + 1
                ...     for i in range(band_count):
                ...         band_name = "band_{}_".format(i + 1)
                ...         builder.add_parameter(name=band_name + "pregain", value=0)
                ...         builder.add_parameter(name=band_name + "clamp_time", value=0.01)
                ...         builder.add_parameter(name=band_name + "relax_time", value=0.1)
                ...         builder.add_parameter(name=band_name + "threshold", value=-6)
                ...         builder.add_parameter(name=band_name + "slope_above", value=0.5)
                ...         builder.add_parameter(name=band_name + "slope_below", value=1.0)
                ...         builder.add_parameter(name=band_name + "postgain", value=0)
                ...

            ::

                >>> def signal_block(builder, source, state):
                ...     bands = []
                ...     frequencies = state["frequencies"]
                ...     for frequency in frequencies:
                ...         band = LPF.ar(source=source, frequency=frequency)
                ...         bands.append(band)
                ...         source -= band
                ...     bands.append(source)
                ...     compressors = []
                ...     for i, band in enumerate(bands):
                ...         band_name = "band_{}_".format(i + 1)
                ...         band *= builder[band_name + "pregain"].db_to_amplitude()
                ...         band = CompanderD.ar(
                ...             source=band,
                ...             clamp_time=builder[band_name + "clamp_time"],
                ...             relax_time=builder[band_name + "relax_time"],
                ...             slope_above=builder[band_name + "slope_above"],
                ...             slope_below=builder[band_name + "slope_below"],
                ...             threshold=builder[band_name + "threshold"].db_to_amplitude(),
                ...         )
                ...         band *= builder[band_name + "postgain"].db_to_amplitude()
                ...         compressors.extend(band if isinstance(band, Sequence) else [band])
                ...     source = Mix.multichannel(
                ...         compressors,
                ...         state["channel_count"],
                ...     )
                ...     return source
                ...

            ::

                >>> from supriya.ugens.factories import SynthDefFactory
                >>> factory = SynthDefFactory()
                >>> factory = factory.with_initial_state(frequencies=(300, 1200, 9600))
                >>> factory = factory.with_parameter_block(parameter_block)
                >>> factory = factory.with_input()
                >>> factory = factory.with_signal_block(signal_block)
                >>> factory = factory.with_output(crossfaded=True)
                >>> synthdef = factory.build()
                >>> supriya.graph(synthdef)  # doctest: +SKIP

            ::

                >>> print(synthdef)
                synthdef:
                    name: ...
                    ugens:
                    -   Control.kr:
                            band_1_clamp_time: 0.01
                            band_1_postgain: 0.0
                            band_1_pregain: 0.0
                            band_1_relax_time: 0.1
                            band_1_slope_above: 0.5
                            band_1_slope_below: 1.0
                            band_1_threshold: -6.0
                            band_2_clamp_time: 0.01
                            band_2_postgain: 0.0
                            band_2_pregain: 0.0
                            band_2_relax_time: 0.1
                            band_2_slope_above: 0.5
                            band_2_slope_below: 1.0
                            band_2_threshold: -6.0
                            band_3_clamp_time: 0.01
                            band_3_postgain: 0.0
                            band_3_pregain: 0.0
                            band_3_relax_time: 0.1
                            band_3_slope_above: 0.5
                            band_3_slope_below: 1.0
                            band_3_threshold: -6.0
                            band_4_clamp_time: 0.01
                            band_4_postgain: 0.0
                            band_4_pregain: 0.0
                            band_4_relax_time: 0.1
                            band_4_slope_above: 0.5
                            band_4_slope_below: 1.0
                            band_4_threshold: -6.0
                            mix: 0.0
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/0:
                            source: Control.kr[2:band_1_pregain]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/1:
                            source: Control.kr[6:band_1_threshold]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/2:
                            source: Control.kr[1:band_1_postgain]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/3:
                            source: Control.kr[9:band_2_pregain]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/4:
                            source: Control.kr[13:band_2_threshold]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/5:
                            source: Control.kr[8:band_2_postgain]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/6:
                            source: Control.kr[16:band_3_pregain]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/7:
                            source: Control.kr[20:band_3_threshold]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/8:
                            source: Control.kr[15:band_3_postgain]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/9:
                            source: Control.kr[23:band_4_pregain]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/10:
                            source: Control.kr[27:band_4_threshold]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/11:
                            source: Control.kr[22:band_4_postgain]
                    -   Control.ir:
                            out: 0.0
                    -   In.ar:
                            bus: Control.ir[0:out]
                    -   LPF.ar/0:
                            source: In.ar[0]
                            frequency: 300.0
                    -   BinaryOpUGen(SUBTRACTION).ar/0:
                            left: In.ar[0]
                            right: LPF.ar/0[0]
                    -   LPF.ar/1:
                            source: BinaryOpUGen(SUBTRACTION).ar/0[0]
                            frequency: 1200.0
                    -   BinaryOpUGen(SUBTRACTION).ar/1:
                            left: BinaryOpUGen(SUBTRACTION).ar/0[0]
                            right: LPF.ar/1[0]
                    -   LPF.ar/2:
                            source: BinaryOpUGen(SUBTRACTION).ar/1[0]
                            frequency: 9600.0
                    -   BinaryOpUGen(SUBTRACTION).ar/2:
                            left: BinaryOpUGen(SUBTRACTION).ar/1[0]
                            right: LPF.ar/2[0]
                    -   BinaryOpUGen(MULTIPLICATION).ar/0:
                            left: BinaryOpUGen(SUBTRACTION).ar/2[0]
                            right: UnaryOpUGen(DB_TO_AMPLITUDE).kr/9[0]
                    -   DelayN.ar/0:
                            source: BinaryOpUGen(MULTIPLICATION).ar/0[0]
                            maximum_delay_time: Control.kr[21:band_4_clamp_time]
                            delay_time: Control.kr[21:band_4_clamp_time]
                    -   Compander.ar/0:
                            source: BinaryOpUGen(MULTIPLICATION).ar/0[0]
                            control: DelayN.ar/0[0]
                            threshold: UnaryOpUGen(DB_TO_AMPLITUDE).kr/10[0]
                            slope_below: Control.kr[26:band_4_slope_below]
                            slope_above: Control.kr[25:band_4_slope_above]
                            clamp_time: Control.kr[21:band_4_clamp_time]
                            relax_time: Control.kr[24:band_4_relax_time]
                    -   BinaryOpUGen(MULTIPLICATION).ar/1:
                            left: Compander.ar/0[0]
                            right: UnaryOpUGen(DB_TO_AMPLITUDE).kr/11[0]
                    -   BinaryOpUGen(MULTIPLICATION).ar/2:
                            left: LPF.ar/2[0]
                            right: UnaryOpUGen(DB_TO_AMPLITUDE).kr/6[0]
                    -   DelayN.ar/1:
                            source: BinaryOpUGen(MULTIPLICATION).ar/2[0]
                            maximum_delay_time: Control.kr[14:band_3_clamp_time]
                            delay_time: Control.kr[14:band_3_clamp_time]
                    -   Compander.ar/1:
                            source: BinaryOpUGen(MULTIPLICATION).ar/2[0]
                            control: DelayN.ar/1[0]
                            threshold: UnaryOpUGen(DB_TO_AMPLITUDE).kr/7[0]
                            slope_below: Control.kr[19:band_3_slope_below]
                            slope_above: Control.kr[18:band_3_slope_above]
                            clamp_time: Control.kr[14:band_3_clamp_time]
                            relax_time: Control.kr[17:band_3_relax_time]
                    -   BinaryOpUGen(MULTIPLICATION).ar/3:
                            left: Compander.ar/1[0]
                            right: UnaryOpUGen(DB_TO_AMPLITUDE).kr/8[0]
                    -   BinaryOpUGen(MULTIPLICATION).ar/4:
                            left: LPF.ar/1[0]
                            right: UnaryOpUGen(DB_TO_AMPLITUDE).kr/3[0]
                    -   DelayN.ar/2:
                            source: BinaryOpUGen(MULTIPLICATION).ar/4[0]
                            maximum_delay_time: Control.kr[7:band_2_clamp_time]
                            delay_time: Control.kr[7:band_2_clamp_time]
                    -   Compander.ar/2:
                            source: BinaryOpUGen(MULTIPLICATION).ar/4[0]
                            control: DelayN.ar/2[0]
                            threshold: UnaryOpUGen(DB_TO_AMPLITUDE).kr/4[0]
                            slope_below: Control.kr[12:band_2_slope_below]
                            slope_above: Control.kr[11:band_2_slope_above]
                            clamp_time: Control.kr[7:band_2_clamp_time]
                            relax_time: Control.kr[10:band_2_relax_time]
                    -   BinaryOpUGen(MULTIPLICATION).ar/5:
                            left: Compander.ar/2[0]
                            right: UnaryOpUGen(DB_TO_AMPLITUDE).kr/5[0]
                    -   BinaryOpUGen(MULTIPLICATION).ar/6:
                            left: LPF.ar/0[0]
                            right: UnaryOpUGen(DB_TO_AMPLITUDE).kr/0[0]
                    -   DelayN.ar/3:
                            source: BinaryOpUGen(MULTIPLICATION).ar/6[0]
                            maximum_delay_time: Control.kr[0:band_1_clamp_time]
                            delay_time: Control.kr[0:band_1_clamp_time]
                    -   Compander.ar/3:
                            source: BinaryOpUGen(MULTIPLICATION).ar/6[0]
                            control: DelayN.ar/3[0]
                            threshold: UnaryOpUGen(DB_TO_AMPLITUDE).kr/1[0]
                            slope_below: Control.kr[5:band_1_slope_below]
                            slope_above: Control.kr[4:band_1_slope_above]
                            clamp_time: Control.kr[0:band_1_clamp_time]
                            relax_time: Control.kr[3:band_1_relax_time]
                    -   BinaryOpUGen(MULTIPLICATION).ar/7:
                            left: Compander.ar/3[0]
                            right: UnaryOpUGen(DB_TO_AMPLITUDE).kr/2[0]
                    -   Sum4.ar:
                            input_one: BinaryOpUGen(MULTIPLICATION).ar/7[0]
                            input_two: BinaryOpUGen(MULTIPLICATION).ar/5[0]
                            input_three: BinaryOpUGen(MULTIPLICATION).ar/3[0]
                            input_four: BinaryOpUGen(MULTIPLICATION).ar/1[0]
                    -   XOut.ar:
                            bus: Control.ir[0:out]
                            crossfade: Control.kr[28:mix]
                            source[0]: Sum4.ar[0]

        .. container:: example

            The above factory can be parameterized to use smaller or larger numbers of
            bands during the build stage:

            ::

                >>> frequencies = (150, 300, 600, 1200, 2400, 4800, 9600)
                >>> synthdef = factory.build(frequencies=frequencies)
                >>> supriya.graph(synthdef)  # doctest: +SKIP

            ::

                >>> print(synthdef)
                synthdef:
                    name: ...
                    ugens:
                    -   Control.kr:
                            band_1_clamp_time: 0.01
                            band_1_postgain: 0.0
                            band_1_pregain: 0.0
                            band_1_relax_time: 0.1
                            band_1_slope_above: 0.5
                            band_1_slope_below: 1.0
                            band_1_threshold: -6.0
                            band_2_clamp_time: 0.01
                            band_2_postgain: 0.0
                            band_2_pregain: 0.0
                            band_2_relax_time: 0.1
                            band_2_slope_above: 0.5
                            band_2_slope_below: 1.0
                            band_2_threshold: -6.0
                            band_3_clamp_time: 0.01
                            band_3_postgain: 0.0
                            band_3_pregain: 0.0
                            band_3_relax_time: 0.1
                            band_3_slope_above: 0.5
                            band_3_slope_below: 1.0
                            band_3_threshold: -6.0
                            band_4_clamp_time: 0.01
                            band_4_postgain: 0.0
                            band_4_pregain: 0.0
                            band_4_relax_time: 0.1
                            band_4_slope_above: 0.5
                            band_4_slope_below: 1.0
                            band_4_threshold: -6.0
                            band_5_clamp_time: 0.01
                            band_5_postgain: 0.0
                            band_5_pregain: 0.0
                            band_5_relax_time: 0.1
                            band_5_slope_above: 0.5
                            band_5_slope_below: 1.0
                            band_5_threshold: -6.0
                            band_6_clamp_time: 0.01
                            band_6_postgain: 0.0
                            band_6_pregain: 0.0
                            band_6_relax_time: 0.1
                            band_6_slope_above: 0.5
                            band_6_slope_below: 1.0
                            band_6_threshold: -6.0
                            band_7_clamp_time: 0.01
                            band_7_postgain: 0.0
                            band_7_pregain: 0.0
                            band_7_relax_time: 0.1
                            band_7_slope_above: 0.5
                            band_7_slope_below: 1.0
                            band_7_threshold: -6.0
                            band_8_clamp_time: 0.01
                            band_8_postgain: 0.0
                            band_8_pregain: 0.0
                            band_8_relax_time: 0.1
                            band_8_slope_above: 0.5
                            band_8_slope_below: 1.0
                            band_8_threshold: -6.0
                            mix: 0.0
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/0:
                            source: Control.kr[2:band_1_pregain]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/1:
                            source: Control.kr[6:band_1_threshold]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/2:
                            source: Control.kr[1:band_1_postgain]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/3:
                            source: Control.kr[9:band_2_pregain]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/4:
                            source: Control.kr[13:band_2_threshold]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/5:
                            source: Control.kr[8:band_2_postgain]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/6:
                            source: Control.kr[16:band_3_pregain]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/7:
                            source: Control.kr[20:band_3_threshold]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/8:
                            source: Control.kr[15:band_3_postgain]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/9:
                            source: Control.kr[23:band_4_pregain]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/10:
                            source: Control.kr[27:band_4_threshold]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/11:
                            source: Control.kr[22:band_4_postgain]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/12:
                            source: Control.kr[30:band_5_pregain]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/13:
                            source: Control.kr[34:band_5_threshold]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/14:
                            source: Control.kr[29:band_5_postgain]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/15:
                            source: Control.kr[37:band_6_pregain]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/16:
                            source: Control.kr[41:band_6_threshold]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/17:
                            source: Control.kr[36:band_6_postgain]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/18:
                            source: Control.kr[44:band_7_pregain]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/19:
                            source: Control.kr[48:band_7_threshold]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/20:
                            source: Control.kr[43:band_7_postgain]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/21:
                            source: Control.kr[51:band_8_pregain]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/22:
                            source: Control.kr[55:band_8_threshold]
                    -   UnaryOpUGen(DB_TO_AMPLITUDE).kr/23:
                            source: Control.kr[50:band_8_postgain]
                    -   Control.ir:
                            out: 0.0
                    -   In.ar:
                            bus: Control.ir[0:out]
                    -   LPF.ar/0:
                            source: In.ar[0]
                            frequency: 150.0
                    -   BinaryOpUGen(SUBTRACTION).ar/0:
                            left: In.ar[0]
                            right: LPF.ar/0[0]
                    -   LPF.ar/1:
                            source: BinaryOpUGen(SUBTRACTION).ar/0[0]
                            frequency: 300.0
                    -   BinaryOpUGen(SUBTRACTION).ar/1:
                            left: BinaryOpUGen(SUBTRACTION).ar/0[0]
                            right: LPF.ar/1[0]
                    -   LPF.ar/2:
                            source: BinaryOpUGen(SUBTRACTION).ar/1[0]
                            frequency: 600.0
                    -   BinaryOpUGen(SUBTRACTION).ar/2:
                            left: BinaryOpUGen(SUBTRACTION).ar/1[0]
                            right: LPF.ar/2[0]
                    -   LPF.ar/3:
                            source: BinaryOpUGen(SUBTRACTION).ar/2[0]
                            frequency: 1200.0
                    -   BinaryOpUGen(SUBTRACTION).ar/3:
                            left: BinaryOpUGen(SUBTRACTION).ar/2[0]
                            right: LPF.ar/3[0]
                    -   LPF.ar/4:
                            source: BinaryOpUGen(SUBTRACTION).ar/3[0]
                            frequency: 2400.0
                    -   BinaryOpUGen(SUBTRACTION).ar/4:
                            left: BinaryOpUGen(SUBTRACTION).ar/3[0]
                            right: LPF.ar/4[0]
                    -   LPF.ar/5:
                            source: BinaryOpUGen(SUBTRACTION).ar/4[0]
                            frequency: 4800.0
                    -   BinaryOpUGen(SUBTRACTION).ar/5:
                            left: BinaryOpUGen(SUBTRACTION).ar/4[0]
                            right: LPF.ar/5[0]
                    -   LPF.ar/6:
                            source: BinaryOpUGen(SUBTRACTION).ar/5[0]
                            frequency: 9600.0
                    -   BinaryOpUGen(SUBTRACTION).ar/6:
                            left: BinaryOpUGen(SUBTRACTION).ar/5[0]
                            right: LPF.ar/6[0]
                    -   BinaryOpUGen(MULTIPLICATION).ar/0:
                            left: BinaryOpUGen(SUBTRACTION).ar/6[0]
                            right: UnaryOpUGen(DB_TO_AMPLITUDE).kr/21[0]
                    -   DelayN.ar/0:
                            source: BinaryOpUGen(MULTIPLICATION).ar/0[0]
                            maximum_delay_time: Control.kr[49:band_8_clamp_time]
                            delay_time: Control.kr[49:band_8_clamp_time]
                    -   Compander.ar/0:
                            source: BinaryOpUGen(MULTIPLICATION).ar/0[0]
                            control: DelayN.ar/0[0]
                            threshold: UnaryOpUGen(DB_TO_AMPLITUDE).kr/22[0]
                            slope_below: Control.kr[54:band_8_slope_below]
                            slope_above: Control.kr[53:band_8_slope_above]
                            clamp_time: Control.kr[49:band_8_clamp_time]
                            relax_time: Control.kr[52:band_8_relax_time]
                    -   BinaryOpUGen(MULTIPLICATION).ar/1:
                            left: Compander.ar/0[0]
                            right: UnaryOpUGen(DB_TO_AMPLITUDE).kr/23[0]
                    -   BinaryOpUGen(MULTIPLICATION).ar/2:
                            left: LPF.ar/6[0]
                            right: UnaryOpUGen(DB_TO_AMPLITUDE).kr/18[0]
                    -   DelayN.ar/1:
                            source: BinaryOpUGen(MULTIPLICATION).ar/2[0]
                            maximum_delay_time: Control.kr[42:band_7_clamp_time]
                            delay_time: Control.kr[42:band_7_clamp_time]
                    -   Compander.ar/1:
                            source: BinaryOpUGen(MULTIPLICATION).ar/2[0]
                            control: DelayN.ar/1[0]
                            threshold: UnaryOpUGen(DB_TO_AMPLITUDE).kr/19[0]
                            slope_below: Control.kr[47:band_7_slope_below]
                            slope_above: Control.kr[46:band_7_slope_above]
                            clamp_time: Control.kr[42:band_7_clamp_time]
                            relax_time: Control.kr[45:band_7_relax_time]
                    -   BinaryOpUGen(MULTIPLICATION).ar/3:
                            left: Compander.ar/1[0]
                            right: UnaryOpUGen(DB_TO_AMPLITUDE).kr/20[0]
                    -   BinaryOpUGen(MULTIPLICATION).ar/4:
                            left: LPF.ar/5[0]
                            right: UnaryOpUGen(DB_TO_AMPLITUDE).kr/15[0]
                    -   DelayN.ar/2:
                            source: BinaryOpUGen(MULTIPLICATION).ar/4[0]
                            maximum_delay_time: Control.kr[35:band_6_clamp_time]
                            delay_time: Control.kr[35:band_6_clamp_time]
                    -   Compander.ar/2:
                            source: BinaryOpUGen(MULTIPLICATION).ar/4[0]
                            control: DelayN.ar/2[0]
                            threshold: UnaryOpUGen(DB_TO_AMPLITUDE).kr/16[0]
                            slope_below: Control.kr[40:band_6_slope_below]
                            slope_above: Control.kr[39:band_6_slope_above]
                            clamp_time: Control.kr[35:band_6_clamp_time]
                            relax_time: Control.kr[38:band_6_relax_time]
                    -   BinaryOpUGen(MULTIPLICATION).ar/5:
                            left: Compander.ar/2[0]
                            right: UnaryOpUGen(DB_TO_AMPLITUDE).kr/17[0]
                    -   BinaryOpUGen(MULTIPLICATION).ar/6:
                            left: LPF.ar/4[0]
                            right: UnaryOpUGen(DB_TO_AMPLITUDE).kr/12[0]
                    -   DelayN.ar/3:
                            source: BinaryOpUGen(MULTIPLICATION).ar/6[0]
                            maximum_delay_time: Control.kr[28:band_5_clamp_time]
                            delay_time: Control.kr[28:band_5_clamp_time]
                    -   Compander.ar/3:
                            source: BinaryOpUGen(MULTIPLICATION).ar/6[0]
                            control: DelayN.ar/3[0]
                            threshold: UnaryOpUGen(DB_TO_AMPLITUDE).kr/13[0]
                            slope_below: Control.kr[33:band_5_slope_below]
                            slope_above: Control.kr[32:band_5_slope_above]
                            clamp_time: Control.kr[28:band_5_clamp_time]
                            relax_time: Control.kr[31:band_5_relax_time]
                    -   BinaryOpUGen(MULTIPLICATION).ar/7:
                            left: Compander.ar/3[0]
                            right: UnaryOpUGen(DB_TO_AMPLITUDE).kr/14[0]
                    -   Sum4.ar/0:
                            input_one: BinaryOpUGen(MULTIPLICATION).ar/7[0]
                            input_two: BinaryOpUGen(MULTIPLICATION).ar/5[0]
                            input_three: BinaryOpUGen(MULTIPLICATION).ar/3[0]
                            input_four: BinaryOpUGen(MULTIPLICATION).ar/1[0]
                    -   BinaryOpUGen(MULTIPLICATION).ar/8:
                            left: LPF.ar/3[0]
                            right: UnaryOpUGen(DB_TO_AMPLITUDE).kr/9[0]
                    -   DelayN.ar/4:
                            source: BinaryOpUGen(MULTIPLICATION).ar/8[0]
                            maximum_delay_time: Control.kr[21:band_4_clamp_time]
                            delay_time: Control.kr[21:band_4_clamp_time]
                    -   Compander.ar/4:
                            source: BinaryOpUGen(MULTIPLICATION).ar/8[0]
                            control: DelayN.ar/4[0]
                            threshold: UnaryOpUGen(DB_TO_AMPLITUDE).kr/10[0]
                            slope_below: Control.kr[26:band_4_slope_below]
                            slope_above: Control.kr[25:band_4_slope_above]
                            clamp_time: Control.kr[21:band_4_clamp_time]
                            relax_time: Control.kr[24:band_4_relax_time]
                    -   BinaryOpUGen(MULTIPLICATION).ar/9:
                            left: Compander.ar/4[0]
                            right: UnaryOpUGen(DB_TO_AMPLITUDE).kr/11[0]
                    -   BinaryOpUGen(MULTIPLICATION).ar/10:
                            left: LPF.ar/2[0]
                            right: UnaryOpUGen(DB_TO_AMPLITUDE).kr/6[0]
                    -   DelayN.ar/5:
                            source: BinaryOpUGen(MULTIPLICATION).ar/10[0]
                            maximum_delay_time: Control.kr[14:band_3_clamp_time]
                            delay_time: Control.kr[14:band_3_clamp_time]
                    -   Compander.ar/5:
                            source: BinaryOpUGen(MULTIPLICATION).ar/10[0]
                            control: DelayN.ar/5[0]
                            threshold: UnaryOpUGen(DB_TO_AMPLITUDE).kr/7[0]
                            slope_below: Control.kr[19:band_3_slope_below]
                            slope_above: Control.kr[18:band_3_slope_above]
                            clamp_time: Control.kr[14:band_3_clamp_time]
                            relax_time: Control.kr[17:band_3_relax_time]
                    -   BinaryOpUGen(MULTIPLICATION).ar/11:
                            left: Compander.ar/5[0]
                            right: UnaryOpUGen(DB_TO_AMPLITUDE).kr/8[0]
                    -   BinaryOpUGen(MULTIPLICATION).ar/12:
                            left: LPF.ar/1[0]
                            right: UnaryOpUGen(DB_TO_AMPLITUDE).kr/3[0]
                    -   DelayN.ar/6:
                            source: BinaryOpUGen(MULTIPLICATION).ar/12[0]
                            maximum_delay_time: Control.kr[7:band_2_clamp_time]
                            delay_time: Control.kr[7:band_2_clamp_time]
                    -   Compander.ar/6:
                            source: BinaryOpUGen(MULTIPLICATION).ar/12[0]
                            control: DelayN.ar/6[0]
                            threshold: UnaryOpUGen(DB_TO_AMPLITUDE).kr/4[0]
                            slope_below: Control.kr[12:band_2_slope_below]
                            slope_above: Control.kr[11:band_2_slope_above]
                            clamp_time: Control.kr[7:band_2_clamp_time]
                            relax_time: Control.kr[10:band_2_relax_time]
                    -   BinaryOpUGen(MULTIPLICATION).ar/13:
                            left: Compander.ar/6[0]
                            right: UnaryOpUGen(DB_TO_AMPLITUDE).kr/5[0]
                    -   BinaryOpUGen(MULTIPLICATION).ar/14:
                            left: LPF.ar/0[0]
                            right: UnaryOpUGen(DB_TO_AMPLITUDE).kr/0[0]
                    -   DelayN.ar/7:
                            source: BinaryOpUGen(MULTIPLICATION).ar/14[0]
                            maximum_delay_time: Control.kr[0:band_1_clamp_time]
                            delay_time: Control.kr[0:band_1_clamp_time]
                    -   Compander.ar/7:
                            source: BinaryOpUGen(MULTIPLICATION).ar/14[0]
                            control: DelayN.ar/7[0]
                            threshold: UnaryOpUGen(DB_TO_AMPLITUDE).kr/1[0]
                            slope_below: Control.kr[5:band_1_slope_below]
                            slope_above: Control.kr[4:band_1_slope_above]
                            clamp_time: Control.kr[0:band_1_clamp_time]
                            relax_time: Control.kr[3:band_1_relax_time]
                    -   BinaryOpUGen(MULTIPLICATION).ar/15:
                            left: Compander.ar/7[0]
                            right: UnaryOpUGen(DB_TO_AMPLITUDE).kr/2[0]
                    -   Sum4.ar/1:
                            input_one: BinaryOpUGen(MULTIPLICATION).ar/15[0]
                            input_two: BinaryOpUGen(MULTIPLICATION).ar/13[0]
                            input_three: BinaryOpUGen(MULTIPLICATION).ar/11[0]
                            input_four: BinaryOpUGen(MULTIPLICATION).ar/9[0]
                    -   BinaryOpUGen(ADDITION).ar:
                            left: Sum4.ar/1[0]
                            right: Sum4.ar/0[0]
                    -   XOut.ar:
                            bus: Control.ir[0:out]
                            crossfade: Control.kr[56:mix]
                            source[0]: BinaryOpUGen(ADDITION).ar[0]

        """
        clone = self._clone()
        clone._parameter_blocks.append(block_function)
        return clone

    def with_rand_id(self, rand_id=0):
        """
        Return a new factory configured with a random number generator ID.

        .. container:: example

            ::

                >>> from supriya.ugens.factories import SynthDefFactory
                >>> factory = SynthDefFactory()

            ::

                >>> def signal_block(builder, source, state):
                ...     iterations = state.get("iterations") or 2
                ...     for _ in range(iterations):
                ...         source = supriya.ugens.AllpassC.ar(
                ...             decay_time=supriya.ugens.ExpRand.ir(minimum=0.01, maximum=0.1),
                ...             delay_time=supriya.ugens.ExpRand.ir(minimum=0.01, maximum=0.1),
                ...             source=source,
                ...             maximum_delay_time=0.1,
                ...         )
                ...     return source

            ::

                >>> factory = factory.with_input()
                >>> factory = factory.with_output()
                >>> factory = factory.with_signal_block(signal_block)

        .. container:: example

            Configure the factory with a random number generator ID, defaulting to 23:

            ::

                >>> factory = factory.with_rand_id(rand_id=23)
                >>> synthdef = factory.build()
                >>> supriya.graph(synthdef)  # doctest: +SKIP

            ::

                >>> print(synthdef)
                synthdef:
                    name: ...
                    ugens:
                    -   Control.ir:
                            out: 0.0
                            rand_id: 23.0
                    -   RandID.ir:
                            rand_id: Control.ir[1:rand_id]
                    -   In.ar:
                            bus: Control.ir[0:out]
                    -   ExpRand.ir/0:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/1:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/0:
                            source: In.ar[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/1[0]
                            decay_time: ExpRand.ir/0[0]
                    -   ExpRand.ir/2:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/3:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/1:
                            source: AllpassC.ar/0[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/3[0]
                            decay_time: ExpRand.ir/2[0]
                    -   Out.ar:
                            bus: Control.ir[0:out]
                            source[0]: AllpassC.ar/1[0]
        """
        clone = self._clone()
        if rand_id is not None:
            clone._rand_id = int(rand_id)
        else:
            clone._rand_id = None
        return clone

    def with_signal_block(self, block_function):
        """
        Return a new factory configured with an additional signal block function.

        Signal block functions take three parameters:

        builder
            the SynthDef builder instance

        source
            a UGenOperable representing the signal flow

        state
            a dictionary of arbitrary key/value pairs for parameterizing the signal and
            parameter block functions

        Signal block functions should return a UGenOperable.

        .. container:: example

            ::

                >>> from supriya.ugens.factories import SynthDefFactory
                >>> factory = SynthDefFactory()

            ::

                >>> def signal_block(builder, source, state):
                ...     iterations = state.get("iterations") or 2
                ...     for _ in range(iterations):
                ...         source = supriya.ugens.AllpassC.ar(
                ...             decay_time=supriya.ugens.ExpRand.ir(minimum=0.01, maximum=0.1),
                ...             delay_time=supriya.ugens.ExpRand.ir(minimum=0.01, maximum=0.1),
                ...             source=source,
                ...             maximum_delay_time=0.1,
                ...         )
                ...     return source

            ::

                >>> factory = factory.with_input()
                >>> factory = factory.with_output()
                >>> factory = factory.with_signal_block(signal_block)

        .. container:: example

            Configure the factory with an additional signal block:

            ::

                >>> def signal_block_post(builder, source, state):
                ...     source = supriya.ugens.LeakDC.ar(source=source)
                ...     source = supriya.ugens.Limiter.ar(
                ...         duration=supriya.ugens.Rand.ir(minimum=0.005, maximum=0.015),
                ...         source=source,
                ...     )
                ...     return source

            ::

                >>> factory = factory.with_signal_block(signal_block_post)
                >>> synthdef = factory.build()
                >>> supriya.graph(synthdef)  # doctest: +SKIP

            ::

                >>> print(synthdef)
                synthdef:
                    name: ...
                    ugens:
                    -   Control.ir:
                            out: 0.0
                    -   In.ar:
                            bus: Control.ir[0:out]
                    -   ExpRand.ir/0:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/1:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/0:
                            source: In.ar[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/1[0]
                            decay_time: ExpRand.ir/0[0]
                    -   ExpRand.ir/2:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/3:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/1:
                            source: AllpassC.ar/0[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/3[0]
                            decay_time: ExpRand.ir/2[0]
                    -   LeakDC.ar:
                            source: AllpassC.ar/1[0]
                            coefficient: 0.995
                    -   Rand.ir:
                            minimum: 0.005
                            maximum: 0.015
                    -   Limiter.ar:
                            source: LeakDC.ar[0]
                            level: 1.0
                            duration: Rand.ir[0]
                    -   Out.ar:
                            bus: Control.ir[0:out]
                            source[0]: Limiter.ar[0]
        """
        clone = self._clone()
        clone._signal_blocks.append(block_function)
        return clone

    def with_silence_detection(self):
        """
        Return a new factory configured with silence detection.

        .. container:: example

            ::

                >>> from supriya.ugens.factories import SynthDefFactory
                >>> factory = SynthDefFactory()

            ::

                >>> def signal_block(builder, source, state):
                ...     iterations = state.get("iterations") or 2
                ...     for _ in range(iterations):
                ...         source = supriya.ugens.AllpassC.ar(
                ...             decay_time=supriya.ugens.ExpRand.ir(minimum=0.01, maximum=0.1),
                ...             delay_time=supriya.ugens.ExpRand.ir(minimum=0.01, maximum=0.1),
                ...             source=source,
                ...             maximum_delay_time=0.1,
                ...         )
                ...     return source

            ::

                >>> factory = factory.with_input()
                >>> factory = factory.with_output()
                >>> factory = factory.with_signal_block(signal_block)

        .. container:: example

            Configure the factory with silence detection.

            ::

                >>> factory = factory.with_silence_detection()
                >>> synthdef = factory.build()
                >>> supriya.graph(synthdef)  # doctest: +SKIP

            ::

                >>> print(synthdef)
                synthdef:
                    name: ...
                    ugens:
                    -   Control.ir:
                            out: 0.0
                    -   In.ar:
                            bus: Control.ir[0:out]
                    -   ExpRand.ir/0:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/1:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/0:
                            source: In.ar[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/1[0]
                            decay_time: ExpRand.ir/0[0]
                    -   ExpRand.ir/2:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/3:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/1:
                            source: AllpassC.ar/0[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/3[0]
                            decay_time: ExpRand.ir/2[0]
                    -   Out.ar:
                            bus: Control.ir[0:out]
                            source[0]: AllpassC.ar/1[0]
                    -   DetectSilence.kr:
                            source: AllpassC.ar/1[0]
                            threshold: 0.0001
                            time: 0.1
                            done_action: 2.0

        .. container:: example

            Silence detection is applied before any output leveling or windowing.

            ::

                >>> factory = factory.with_output(
                ...     leveled=True,
                ...     windowed=True,
                ... )
                >>> synthdef = factory.build()
                >>> supriya.graph(synthdef)  # doctest: +SKIP

            ::

                >>> print(synthdef)
                synthdef:
                    name: ...
                    ugens:
                    -   Control.kr:
                            level: 1.0
                    -   Control.ir:
                            duration: 1.0
                            out: 0.0
                    -   Line.kr:
                            start: 0.0
                            stop: 1.0
                            duration: Control.ir[0:duration]
                            done_action: 2.0
                    -   UnaryOpUGen(HANNING_WINDOW).kr:
                            source: Line.kr[0]
                    -   In.ar:
                            bus: Control.ir[1:out]
                    -   ExpRand.ir/0:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/1:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/0:
                            source: In.ar[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/1[0]
                            decay_time: ExpRand.ir/0[0]
                    -   ExpRand.ir/2:
                            minimum: 0.01
                            maximum: 0.1
                    -   ExpRand.ir/3:
                            minimum: 0.01
                            maximum: 0.1
                    -   AllpassC.ar/1:
                            source: AllpassC.ar/0[0]
                            maximum_delay_time: 0.1
                            delay_time: ExpRand.ir/3[0]
                            decay_time: ExpRand.ir/2[0]
                    -   BinaryOpUGen(MULTIPLICATION).ar/0:
                            left: AllpassC.ar/1[0]
                            right: Control.kr[0:level]
                    -   BinaryOpUGen(MULTIPLICATION).ar/1:
                            left: BinaryOpUGen(MULTIPLICATION).ar/0[0]
                            right: UnaryOpUGen(HANNING_WINDOW).kr[0]
                    -   Out.ar:
                            bus: Control.ir[1:out]
                            source[0]: BinaryOpUGen(MULTIPLICATION).ar/1[0]
                    -   DetectSilence.kr:
                            source: AllpassC.ar/1[0]
                            threshold: 0.0001
                            time: 0.1
                            done_action: 2.0
        """
        clone = self._clone()
        clone._silence_detection = True
        return clone



import logging

import pytest

from supriya.contexts.errors import MomentClosed
from supriya.contexts.nonrealtime import Score
from supriya.osc import OscBundle, OscMessage
from supriya.ugens import Out, SinOsc, SynthDefBuilder, compile_synthdefs


@pytest.fixture(autouse=True)
def use_caplog(caplog):
    caplog.set_level(logging.INFO)


@pytest.fixture
def context(request):
    return Score()


@pytest.fixture
def synthdefs():
    with SynthDefBuilder(bus=0, frequency=440) as builder:
        Out.ar(bus=builder["bus"], source=SinOsc.ar(frequency=builder["frequency"]))
    synthdef_a = builder.build(name="synthdef-a")
    synthdef_b = builder.build(name="synthdef-b")
    synthdef_c = builder.build(name="synthdef-c")
    return [synthdef_a, synthdef_b, synthdef_c]


def test_add_synthdefs(context, synthdefs):
    def compiled(*x):
        return compile_synthdefs(*x)

    with context.at(0):
        # no synthdefs provided
        with pytest.raises(ValueError):
            context.add_synthdefs()
        # /d_recv
        context.add_synthdefs(synthdefs[0])
        # multiples
        context.add_synthdefs(*synthdefs)
        # completion via on_completion lambda succeeds
        context.add_synthdefs(synthdefs[1], on_completion=lambda ctx: ctx.add_group())
        # completion without moment errors
        completion = context.add_synthdefs(synthdefs[2])
    with pytest.raises(MomentClosed):
        with completion:
            context.add_group()
    with context.at(1.23):
        # completion inside moment succeeds
        with context.add_synthdefs(synthdefs[2]):
            context.add_group()
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(
            contents=(
                OscMessage("/d_recv", compiled(synthdefs[0])),
                OscMessage("/d_recv", compiled(*synthdefs)),
                OscMessage(
                    "/d_recv",
                    compiled(synthdefs[1]),
                    OscMessage("/g_new", 1000, 0, 0),
                ),
                OscMessage("/d_recv", compiled(synthdefs[2])),
            ),
            timestamp=0.0,
        ),
        OscBundle(
            contents=[
                OscMessage(
                    "/d_recv",
                    compiled(synthdefs[2]),
                    OscMessage("/g_new", 1001, 0, 0),
                )
            ],
            timestamp=1.23,
        ),
    ]


def test_free_synthdefs(context, synthdefs):
    with context.at(0):
        # no synthdefs provided
        with pytest.raises(ValueError):
            context.free_synthdefs()
        # /d_free
        context.free_synthdefs(*synthdefs)
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(
            contents=(OscMessage("/d_free", "synthdef-a", "synthdef-b", "synthdef-c"),),
            timestamp=0.0,
        )
    ]


def test_free_all_synthdefs(context):
    with context.at(0):
        context.free_all_synthdefs()
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(contents=(OscMessage("/d_freeAll"),), timestamp=0.0)
    ]


def test_load_synthdefs(context, tmp_path):
    with context.at(0):
        context.load_synthdefs(tmp_path / "a.scsyndef")
        # completion via on_completion lambda succeeds
        context.load_synthdefs(
            tmp_path / "b.scsyndef", on_completion=lambda ctx: ctx.add_group()
        )
        # completion without moment errors
        completion = context.load_synthdefs(tmp_path / "c.scsyndef")
    with pytest.raises(MomentClosed):
        with completion:
            context.add_group()
    with context.at(1.23):
        # completion inside moment succeeds
        with context.load_synthdefs(tmp_path / "c.scsyndef"):
            context.add_group()
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(
            contents=(
                OscMessage("/d_load", str(tmp_path / "a.scsyndef")),
                OscMessage(
                    "/d_load",
                    str(tmp_path / "b.scsyndef"),
                    OscMessage("/g_new", 1000, 0, 0),
                ),
                OscMessage("/d_load", str(tmp_path / "c.scsyndef")),
            ),
            timestamp=0.0,
        ),
        OscBundle(
            contents=(
                OscMessage(
                    "/d_load",
                    str(tmp_path / "c.scsyndef"),
                    OscMessage("/g_new", 1001, 0, 0),
                ),
            ),
            timestamp=1.23,
        ),
    ]


def test_load_synthdefs_directory(context, tmp_path):
    with context.at(0):
        context.load_synthdefs_directory(tmp_path / "a")
        # completion via on_completion lambda succeeds
        context.load_synthdefs_directory(
            tmp_path / "b", on_completion=lambda ctx: ctx.add_group()
        )
        # completion without moment errors
        completion = context.load_synthdefs_directory(tmp_path / "c")
    with pytest.raises(MomentClosed):
        with completion:
            context.add_group()
    with context.at(1.23):
        # completion inside moment succeeds
        with context.load_synthdefs_directory(tmp_path / "c"):
            context.add_group()
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(
            contents=(
                OscMessage("/d_loadDir", str(tmp_path / "a")),
                OscMessage(
                    "/d_loadDir", str(tmp_path / "b"), OscMessage("/g_new", 1000, 0, 0)
                ),
                OscMessage("/d_loadDir", str(tmp_path / "c")),
            ),
            timestamp=0.0,
        ),
        OscBundle(
            contents=(
                OscMessage(
                    "/d_loadDir", str(tmp_path / "c"), OscMessage("/g_new", 1001, 0, 0)
                ),
            ),
            timestamp=1.23,
        ),
    ]



from typing import Callable, List, Optional, Tuple
from typing import Type as TypingType

from mypy.nodes import ARG_OPT, Argument, AssignmentStmt, CallExpr, RefExpr, Var
from mypy.plugin import ClassDefContext, Plugin
from mypy.plugins.common import _get_bool_argument, _get_decorator_bool_argument
from mypy.types import NoneType


class UGenTransformer:
    def __init__(self, ctx: ClassDefContext) -> None:
        self._ctx = ctx

    def collect_params(self) -> List[Tuple[str, bool]]:
        cls = self._ctx.cls
        params = []
        for stmt in cls.defs.body:
            if not isinstance(stmt, AssignmentStmt):
                continue
            expr = stmt.rvalue
            if (
                isinstance(expr, CallExpr)
                and isinstance(expr.callee, RefExpr)
                and expr.callee.fullname in ["supriya.ugens.core.param"]
            ):
                params.append(
                    (
                        getattr(stmt.lvalues[0], "name", ""),
                        _get_bool_argument(self._ctx, expr, "unexpanded", False),
                    )
                )
        return params

    def transform(self) -> bool:
        # is_classmethod flag is not released yet
        from mypy.plugins.common import add_attribute_to_class, add_method_to_class

        api = self._ctx.api
        cls = self._ctx.cls
        info = self._ctx.cls.info

        SupportsIntType = api.named_type("typing.SupportsInt")
        CalculationRateType = api.named_type("supriya.enums.CalculationRate")
        UGenOperableType = api.named_type("supriya.ugens.core.UGenOperable")
        UGenScalarType = api.named_type("supriya.ugens.core.UGenScalar")
        UGenVectorType = api.named_type("supriya.ugens.core.UGenVector")
        # api.named_type() breaks for these... why?
        UGenRecursiveInputTypeSym = api.lookup_fully_qualified(
            "supriya.ugens.core.UGenRecursiveInput"
        )
        UGenScalarInputTypeSym = api.lookup_fully_qualified(
            "supriya.ugens.core.UGenScalarInput"
        )
        UGenVectorInputTypeSym = api.lookup_fully_qualified(
            "supriya.ugens.core.UGenVectorInput"
        )

        assert UGenRecursiveInputTypeSym.node is not None
        assert UGenScalarInputTypeSym.node is not None
        assert UGenVectorInputTypeSym.node is not None

        UGenRecursiveInputType = getattr(UGenRecursiveInputTypeSym.node, "target")
        UGenScalarInputType = getattr(UGenScalarInputTypeSym.node, "target")
        UGenVectorInputType = getattr(UGenVectorInputTypeSym.node, "target")

        decorator_arguments = {
            "ar": _get_decorator_bool_argument(self._ctx, "ar", False),
            "kr": _get_decorator_bool_argument(self._ctx, "kr", False),
            "ir": _get_decorator_bool_argument(self._ctx, "ir", False),
            "dr": _get_decorator_bool_argument(self._ctx, "dr", False),
            "new": _get_decorator_bool_argument(self._ctx, "new", False),
            "is_multichannel": _get_decorator_bool_argument(
                self._ctx, "is_multichannel", False
            ),
            "fixed_channel_count": _get_decorator_bool_argument(
                self._ctx, "fixed_channel_count", False
            ),
        }
        init_args = [
            Argument(
                variable=Var("calculation_rate", CalculationRateType),
                type_annotation=CalculationRateType,
                initializer=None,
                kind=ARG_OPT,
            ),
            Argument(
                variable=Var("special_index", SupportsIntType),
                type_annotation=SupportsIntType,
                initializer=None,
                kind=ARG_OPT,
            ),
        ]
        rate_args = []
        for name, unexpanded in self.collect_params():
            init_type = UGenVectorInputType if unexpanded else UGenScalarInputType
            init_args.append(
                Argument(
                    variable=Var(name, init_type),
                    type_annotation=init_type,
                    initializer=None,
                    kind=ARG_OPT,
                )
            )
            rate_args.append(
                Argument(
                    variable=Var(name, UGenRecursiveInputType),
                    type_annotation=UGenRecursiveInputType,
                    initializer=None,
                    kind=ARG_OPT,
                )
            )
            add_attribute_to_class(
                api=api,
                cls=cls,
                name=name,
                typ=UGenVectorType if unexpanded else UGenScalarType,
                override_allow_incompatible=True,
            )
        if (
            decorator_arguments["is_multichannel"]
            and not decorator_arguments["fixed_channel_count"]
        ):
            channel_count_arg = Argument(
                variable=Var("channel_count", SupportsIntType),
                type_annotation=SupportsIntType,
                initializer=None,
                kind=ARG_OPT,
            )
            init_args.append(channel_count_arg)
            rate_args.append(channel_count_arg)
        for name in ["ar", "kr", "ir", "dr", "new"]:
            if not decorator_arguments[name] or name in info.names:
                continue
            add_method_to_class(
                api=api,
                cls=cls,
                name=name,
                args=rate_args,
                return_type=UGenOperableType,
                is_classmethod=True,
            )
        if "__init__" not in info.names:
            add_method_to_class(
                api=api,
                cls=cls,
                name="__init__",
                args=init_args,
                return_type=NoneType(),
            )
        return True


def _ugen_hook(ctx: ClassDefContext) -> bool:
    return UGenTransformer(ctx).transform()


class SupriyaPlugin(Plugin):
    def get_class_decorator_hook_2(
        self, fullname: str
    ) -> Optional[Callable[[ClassDefContext], bool]]:
        if fullname == "supriya.ugens.core.ugen":
            return _ugen_hook
        return None


def plugin(version: str) -> TypingType[SupriyaPlugin]:
    return SupriyaPlugin



import math

from ..enums import CalculationRate
from .basic import Mix
from .core import (
    PseudoUGen,
    UGen,
    UGenOperable,
    UGenRecursiveParams,
    UGenVector,
    _get_method_for_rate,
    param,
    ugen,
)


@ugen(ar=True, kr=True, channel_count=2, fixed_channel_count=True)
class Balance2(UGen):
    """
    A stereo signal balancer.

    ::

        >>> left = supriya.ugens.WhiteNoise.ar()
        >>> right = supriya.ugens.SinOsc.ar()
        >>> balance_2 = supriya.ugens.Balance2.ar(
        ...     left=left,
        ...     level=1,
        ...     position=0,
        ...     right=right,
        ... )
        >>> balance_2
        <Balance2.ar()>
    """

    left = param()
    right = param()
    position = param(0.0)
    level = param(1.0)


@ugen(ar=True, kr=True, channel_count=3, fixed_channel_count=True)
class BiPanB2(UGen):
    """
    A 2D ambisonic b-format panner.

    ::

        >>> in_a = supriya.ugens.SinOsc.ar()
        >>> in_b = supriya.ugens.WhiteNoise.ar()
        >>> bi_pan_b_2 = supriya.ugens.BiPanB2.ar(
        ...     azimuth=-0.5,
        ...     gain=1,
        ...     in_a=in_a,
        ...     in_b=in_b,
        ... )
        >>> bi_pan_b_2
        <BiPanB2.ar()>

    ::

        >>> w, x, y = bi_pan_b_2
    """

    in_a = param()
    in_b = param()
    azimuth = param()
    gain = param(1.0)


@ugen(ar=True, kr=True, is_multichannel=True, channel_count=4)
class DecodeB2(UGen):
    """
    A 2D Ambisonic B-format decoder.

    ::

        >>> source = supriya.ugens.PinkNoise.ar()
        >>> w, x, y = supriya.ugens.PanB2.ar(
        ...     source=source,
        ...     azimuth=supriya.ugens.SinOsc.kr(),
        ... )
        >>> decode_b_2 = supriya.ugens.DecodeB2.ar(
        ...     channel_count=4,
        ...     orientation=0.5,
        ...     w=w,
        ...     x=x,
        ...     y=y,
        ... )
        >>> decode_b_2
        <DecodeB2.ar()>
    """

    w = param()
    x = param()
    y = param()
    orientation = param(0.5)


@ugen(ar=True, kr=True, channel_count=2, fixed_channel_count=True)
class Pan2(UGen):
    """
    A two channel equal power panner.

    ::

        >>> source = supriya.ugens.WhiteNoise.ar()
        >>> pan_2 = supriya.ugens.Pan2.ar(
        ...     source=source,
        ... )
        >>> pan_2
        <Pan2.ar()>
    """

    source = param()
    position = param(0.0)
    level = param(1.0)


@ugen(ar=True, kr=True, channel_count=4, fixed_channel_count=True)
class Pan4(UGen):
    """
    A four-channel equal-power panner.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> pan_4 = supriya.ugens.Pan4.ar(
        ...     gain=1,
        ...     source=source,
        ...     x_position=0,
        ...     y_position=0,
        ... )
        >>> pan_4
        <Pan4.ar()>
    """

    source = param()
    x_position = param(0)
    y_position = param(0)
    gain = param(1)


@ugen(ar=True, kr=True, is_multichannel=True)
class PanAz(UGen):
    """
    A multi-channel equal-power panner.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> pan_az = supriya.ugens.PanAz.ar(
        ...     channel_count=8,
        ...     amplitude=1,
        ...     orientation=0.5,
        ...     position=0,
        ...     source=source,
        ...     width=2,
        ... )
        >>> pan_az
        <PanAz.ar()>
    """

    source = param()
    position = param(0)
    amplitude = param(1)
    width = param(2)
    orientation = param(0.5)


@ugen(ar=True, kr=True, channel_count=3, fixed_channel_count=True)
class PanB(UGen):
    """
    A 3D ambisonic b-format panner.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> pan_b = supriya.ugens.PanB.ar(
        ...     azimuth=0,
        ...     elevation=0,
        ...     gain=1,
        ...     source=source,
        ... )
        >>> pan_b
        <PanB.ar()>
    """

    source = param()
    azimuth = param(0)
    elevation = param(0)
    gain = param(1)


@ugen(ar=True, kr=True, channel_count=3, fixed_channel_count=True)
class PanB2(UGen):
    """
    A 2D ambisonic b-format panner.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> pan_b_2 = supriya.ugens.PanB2.ar(
        ...     azimuth=0,
        ...     gain=1,
        ...     source=source,
        ... )
        >>> pan_b_2
        <PanB2.ar()>
    """

    source = param()
    azimuth = param(0)
    gain = param(1)


@ugen(ar=True, kr=True, channel_count=2, fixed_channel_count=True)
class Rotate2(UGen):
    """
    Equal-power sound-field rotator.

    ::

        >>> x = supriya.ugens.PinkNoise.ar() * 0.4
        >>> y = supriya.ugens.LFTri.ar(frequency=880)
        >>> y *= supriya.ugens.LFPulse.kr(frequency=3, width=0.1)
        >>> position = supriya.ugens.LFSaw.kr(frequency=0.1)
        >>> rotate_2 = supriya.ugens.Rotate2.ar(
        ...     x=x,
        ...     y=y,
        ...     position=position,
        ... )
        >>> rotate_2
        <Rotate2.ar()>

    Returns an array of the rotator's left and right outputs.
    """

    x = param()
    y = param()
    position = param(0)


class Splay(PseudoUGen):
    """
    A stereo signal spreader.

    ::

        >>> source = supriya.ugens.SinOsc.ar(frequency=[333, 444, 555, 666, 777])
        >>> splay = supriya.ugens.Splay.ar(source=source)
        >>> splay
        <UGenVector([<BinaryOpUGen.ar(MULTIPLICATION)[0]>, <BinaryOpUGen.ar(MULTIPLICATION)[0]>])>

    ::

        >>> print(splay)
        synthdef:
            name: ...
            ugens:
            -   SinOsc.ar/0:
                    frequency: 333.0
                    phase: 0.0
            -   Pan2.ar/0:
                    source: SinOsc.ar/0[0]
                    position: -1.0
                    level: 1.0
            -   SinOsc.ar/1:
                    frequency: 444.0
                    phase: 0.0
            -   Pan2.ar/1:
                    source: SinOsc.ar/1[0]
                    position: -0.5
                    level: 1.0
            -   SinOsc.ar/2:
                    frequency: 555.0
                    phase: 0.0
            -   Pan2.ar/2:
                    source: SinOsc.ar/2[0]
                    position: 0.0
                    level: 1.0
            -   SinOsc.ar/3:
                    frequency: 666.0
                    phase: 0.0
            -   Pan2.ar/3:
                    source: SinOsc.ar/3[0]
                    position: 0.5
                    level: 1.0
            -   Sum4.ar/0:
                    input_one: Pan2.ar/0[0]
                    input_two: Pan2.ar/1[0]
                    input_three: Pan2.ar/2[0]
                    input_four: Pan2.ar/3[0]
            -   Sum4.ar/1:
                    input_one: Pan2.ar/0[1]
                    input_two: Pan2.ar/1[1]
                    input_three: Pan2.ar/2[1]
                    input_four: Pan2.ar/3[1]
            -   SinOsc.ar/4:
                    frequency: 777.0
                    phase: 0.0
            -   Pan2.ar/4:
                    source: SinOsc.ar/4[0]
                    position: 1.0
                    level: 1.0
            -   BinaryOpUGen(ADDITION).ar/0:
                    left: Sum4.ar/0[0]
                    right: Pan2.ar/4[0]
            -   BinaryOpUGen(MULTIPLICATION).ar/0:
                    left: BinaryOpUGen(ADDITION).ar/0[0]
                    right: 0.4472135954999579
            -   BinaryOpUGen(ADDITION).ar/1:
                    left: Sum4.ar/1[0]
                    right: Pan2.ar/4[1]
            -   BinaryOpUGen(MULTIPLICATION).ar/1:
                    left: BinaryOpUGen(ADDITION).ar/1[0]
                    right: 0.4472135954999579
    """

    ### CLASS VARIABLES ###

    _ordered_keys = (
        "spread",
        "level",
        "center",
        "normalize",
        "source",
    )
    _unexpanded_keys = ("source",)

    @classmethod
    def _new_expanded(cls, calculation_rate=None, **kwargs):
        def recurse(
            all_expanded_params: UGenRecursiveParams,
        ) -> UGenOperable:
            if (
                not isinstance(all_expanded_params, dict)
                and len(all_expanded_params) == 1
            ):
                all_expanded_params = all_expanded_params[0]
            if isinstance(all_expanded_params, dict):
                return cls._new_single(
                    calculation_rate=calculation_rate,
                    special_index=0,
                    **all_expanded_params,
                )
            return UGenVector(
                *(recurse(expanded_params) for expanded_params in all_expanded_params)
            )

        return Mix.multichannel(
            recurse(UGen._expand_params(kwargs, unexpanded_keys=cls._unexpanded_keys)),
            2,
        )

    @classmethod
    def _new_single(
        cls,
        *,
        calculation_rate=None,
        center=0,
        level=1,
        normalize=True,
        source=None,
        spread=1,
        **kwargs,
    ):
        positions = [
            (i * (2 / (len(source) - 1)) - 1) * spread + center
            for i in range(len(source))
        ]
        if normalize:
            if calculation_rate == CalculationRate.AUDIO:
                level = level * math.sqrt(1 / len(source))
            else:
                level = level / len(source)
        panners = _get_method_for_rate(Pan2, calculation_rate)(
            source=source, position=positions
        )
        return Mix.multichannel(panners, 2) * level

    @classmethod
    def ar(cls, *, source, center=0, level=1, normalize=True, spread=1):
        return cls._new_expanded(
            calculation_rate=CalculationRate.AUDIO,
            center=center,
            level=level,
            normalize=normalize,
            source=source,
            spread=spread,
        )

    @classmethod
    def kr(cls, *, source, center=0, level=1, normalize=True, spread=1):
        return cls._new_expanded(
            calculation_rate=CalculationRate.CONTROL,
            center=center,
            level=level,
            normalize=normalize,
            source=source,
            spread=spread,
        )


@ugen(ar=True, kr=True)
class XFade2(UGen):
    """
    Two channel equal power crossfader.

    ::

        >>> xfade_3 = supriya.ugens.XFade2.ar(
        ...     in_a=supriya.ugens.Saw.ar(),
        ...     in_b=supriya.ugens.SinOsc.ar(),
        ...     level=1,
        ...     pan=supriya.ugens.LFTri.kr(frequency=0.1),
        ... )
        >>> xfade_3
        <XFade2.ar()[0]>
    """

    in_a = param()
    in_b = param(0)
    pan = param(0)
    level = param(1)



"""
The core pattern classes.
"""

import abc
import inspect
import itertools
import operator
import random
from typing import (
    TYPE_CHECKING,
    Callable,
    Coroutine,
    Dict,
    Generator,
    Generic,
    Iterator,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
)
from uuid import UUID

from uqbar.objects import get_vars

import supriya.patterns
from supriya.clocks import BaseClock, Clock, ClockContext, OfflineClock, Quantization
from supriya.contexts import Bus, Context, Node, Score

from .events import CompositeEvent, Event, Priority

if TYPE_CHECKING:
    from .players import PatternPlayer

T = TypeVar("T")


class Pattern(Generic[T], metaclass=abc.ABCMeta):
    ### CLASSMETHODS ###

    _rngs: Dict[int, Iterator[float]] = {}

    ### SPECIAL METHODS ###

    def __abs__(self) -> "UnaryOpPattern[T]":
        return UnaryOpPattern(operator.abs, self)

    def __add__(self, expr: Union["Pattern[T]", T]) -> "BinaryOpPattern[T]":
        return BinaryOpPattern(operator.add, self, expr)

    def __and__(self, expr: Union["Pattern[T]", T]) -> "BinaryOpPattern[T]":
        return BinaryOpPattern(operator.and_, self, expr)

    def __eq__(self, expr) -> bool:
        self_values = type(self), get_vars(self)
        try:
            expr_values = type(expr), get_vars(expr)
        except AttributeError:
            expr_values = type(expr), expr
        return self_values == expr_values

    def __floordiv__(self, expr: Union["Pattern[T]", T]) -> "BinaryOpPattern[T]":
        return BinaryOpPattern(operator.floordiv, self, expr)

    def __ge__(self, expr: Union["Pattern[T]", T]) -> "BinaryOpPattern[T]":
        return BinaryOpPattern(operator.ge, self, expr)

    def __gt__(self, expr: Union["Pattern[T]", T]) -> "BinaryOpPattern[T]":
        return BinaryOpPattern(operator.gt, self, expr)

    def __invert__(self) -> "UnaryOpPattern[T]":
        return UnaryOpPattern(operator.invert, self)

    def __iter__(self) -> Generator[T, bool, None]:
        should_stop = False
        state: Optional[Dict[str, UUID]] = self._setup_state()
        iterator = self._iterate(state)
        try:
            expr = self._adjust_recursive(next(iterator), state=state)
        except StopIteration:
            return
        start_event, stop_event = self._setup_peripherals(state)
        if start_event:
            should_stop = (yield start_event) or should_stop
        if not should_stop:
            should_stop = (yield expr) or should_stop
            while True:  # Exhaust iterator, even if scheduled to stop
                try:
                    expr = self._adjust_recursive(
                        iterator.send(should_stop), state=state
                    )
                    should_stop = (yield expr) or should_stop
                except StopIteration:
                    break
        if stop_event:
            yield stop_event

    def __le__(self, expr: Union["Pattern[T]", T]) -> "BinaryOpPattern[T]":
        return BinaryOpPattern(operator.le, self, expr)

    def __lshift__(self, expr: Union["Pattern[T]", T]) -> "BinaryOpPattern[T]":
        return BinaryOpPattern(operator.lshift, self, expr)

    def __lt__(self, expr: Union["Pattern[T]", T]) -> "BinaryOpPattern[T]":
        return BinaryOpPattern(operator.lt, self, expr)

    def __mod__(self, expr: Union["Pattern[T]", T]) -> "BinaryOpPattern[T]":
        return BinaryOpPattern(operator.mod, self, expr)

    def __mul__(self, expr: Union["Pattern[T]", T]) -> "BinaryOpPattern[T]":
        return BinaryOpPattern(operator.mul, self, expr)

    def __neg__(self) -> "UnaryOpPattern[T]":
        return UnaryOpPattern(operator.neg, self)

    def __or__(self, expr: Union["Pattern[T]", T]) -> "BinaryOpPattern[T]":
        return BinaryOpPattern(operator.or_, self, expr)

    def __pos__(self) -> "UnaryOpPattern[T]":
        return UnaryOpPattern(operator.pos, self)

    def __pow__(self, expr: Union["Pattern[T]", T]) -> "BinaryOpPattern[T]":
        return BinaryOpPattern(operator.pow, self, expr)

    def __radd__(self, expr: Union["Pattern[T]", T]) -> "BinaryOpPattern[T]":
        return BinaryOpPattern(operator.add, expr, self)

    def __rand__(self, expr: Union["Pattern[T]", T]) -> "BinaryOpPattern[T]":
        return BinaryOpPattern(operator.and_, expr, self)

    def __rfloordiv__(self, expr: Union["Pattern[T]", T]) -> "BinaryOpPattern[T]":
        return BinaryOpPattern(operator.floordiv, expr, self)

    def __rlshift__(self, expr: Union["Pattern[T]", T]) -> "BinaryOpPattern[T]":
        return BinaryOpPattern(operator.lshift, expr, self)

    def __rmod__(self, expr: Union["Pattern[T]", T]) -> "BinaryOpPattern[T]":
        return BinaryOpPattern(operator.mod, expr, self)

    def __rmul__(self, expr: Union["Pattern[T]", T]) -> "BinaryOpPattern[T]":
        return BinaryOpPattern(operator.mul, expr, self)

    def __ror__(self, expr: Union["Pattern[T]", T]) -> "BinaryOpPattern[T]":
        return BinaryOpPattern(operator.or_, expr, self)

    def __rpow__(self, expr: Union["Pattern[T]", T]) -> "BinaryOpPattern[T]":
        return BinaryOpPattern(operator.pow, expr, self)

    def __rrshift__(self, expr: Union["Pattern[T]", T]) -> "BinaryOpPattern[T]":
        return BinaryOpPattern(operator.rshift, expr, self)

    def __rshift__(self, expr: Union["Pattern[T]", T]) -> "BinaryOpPattern[T]":
        return BinaryOpPattern(operator.rshift, self, expr)

    def __rsub__(self, expr: Union["Pattern[T]", T]) -> "BinaryOpPattern[T]":
        return BinaryOpPattern(operator.sub, expr, self)

    def __rtruediv__(self, expr: Union["Pattern[T]", T]) -> "BinaryOpPattern[T]":
        return BinaryOpPattern(operator.truediv, expr, self)

    def __rxor__(self, expr: Union["Pattern[T]", T]) -> "BinaryOpPattern[T]":
        return BinaryOpPattern(operator.xor, expr, self)

    def __sub__(self, expr: Union["Pattern[T]", T]) -> "BinaryOpPattern[T]":
        return BinaryOpPattern(operator.sub, self, expr)

    def __truediv__(self, expr: Union["Pattern[T]", T]) -> "BinaryOpPattern[T]":
        return BinaryOpPattern(operator.truediv, self, expr)

    def __xor__(self, expr: Union["Pattern[T]", T]) -> "BinaryOpPattern[T]":
        return BinaryOpPattern(operator.xor, self, expr)

    ### PRIVATE METHODS ###

    def _adjust(self, expr: T, state: Optional[Dict[str, UUID]] = None) -> T:
        return expr

    def _adjust_recursive(self, expr: T, state: Optional[Dict[str, UUID]] = None) -> T:
        if isinstance(expr, CompositeEvent):
            return cast(
                T,
                CompositeEvent(
                    [
                        cast(Event, self._adjust(cast(T, event), state=state))
                        for event in expr.events
                    ],
                    delta=expr.delta,
                ),
            )
        return self._adjust(expr, state=state)

    def _apply_recursive(self, procedure, *exprs):
        if all(
            not isinstance(x, Sequence) or isinstance(x, (str, bytes)) for x in exprs
        ):
            return procedure(*exprs)
        coerced_exprs = [
            (
                expr
                if ((isinstance(expr, Sequence) and not isinstance(expr, (str, bytes))))
                else [expr]
            )
            for expr in exprs
        ]
        max_length = max(len(expr) for expr in coerced_exprs)
        for i, expr in enumerate(coerced_exprs):
            if len(expr) < max_length:
                cycle = itertools.cycle(expr)
                coerced_exprs[i] = [next(cycle) for _ in range(max_length)]
        return tuple(
            self._apply_recursive(procedure, *items) for items in zip(*coerced_exprs)
        )

    def _freeze_recursive(self, value):
        if isinstance(value, str):
            return value
        elif isinstance(value, Sequence) and not isinstance(value, Pattern):
            return tuple(self._freeze_recursive(_) for _ in value)
        return value

    def _get_rng(self) -> Iterator[float]:
        identifier = None
        try:
            # Walk frames to find an enclosing SeedPattern._iterate()
            frame = inspect.currentframe()
            while frame is not None:
                if (
                    isinstance(frame.f_locals.get("self"), SeedPattern)
                    and frame.f_code.co_name == "_iterate"
                ):
                    identifier = id(frame)
                    break
                frame = frame.f_back
        finally:
            del frame
        if identifier in self._rngs:
            return self._rngs[identifier]
        return self._get_stdlib_rng()

    def _get_seeded_rng(self, seed: int = 1) -> Iterator[float]:
        mask = 0x7FFFFFFF
        while True:
            seed = (seed * 1_103_515_245 + 12345) & mask
            yield float(seed) / mask

    def _get_stdlib_rng(self) -> Iterator[float]:
        while True:
            yield random.random()

    @abc.abstractmethod
    def _iterate(
        self, state: Optional[Dict[str, UUID]] = None
    ) -> Generator[T, bool, None]:
        raise NotImplementedError

    def _loop(self, iterations: Optional[int] = None) -> Iterator[bool]:
        if iterations is None:
            while True:
                yield True
        else:
            for _ in range(iterations):
                yield True

    def _setup_state(self) -> Optional[Dict[str, UUID]]:
        return None

    def _setup_peripherals(
        self, state: Optional[Dict[str, UUID]]
    ) -> Tuple[Optional[T], Optional[T]]:
        return None, None

    ### PUBLIC METHODS ###

    def play(
        self,
        context: Context,
        *,
        at: Optional[float] = None,
        callback: Optional[
            Callable[
                ["supriya.patterns.PatternPlayer", ClockContext, Event, Priority],
                Optional[Coroutine],
            ]
        ] = None,
        clock: Optional[BaseClock] = None,
        quantization: Optional[Quantization] = None,
        target_bus: Optional[Bus] = None,
        target_node: Optional[Node] = None,
        tempo: Optional[float] = None,
        until: Optional[float] = None,
        uuid: Optional[UUID] = None,
    ) -> "PatternPlayer":
        from .players import PatternPlayer  # Avoid circular import

        if isinstance(context, Score):
            clock = OfflineClock()
            at = at or 0.0
        elif clock is None:
            clock = Clock.default()
        player = PatternPlayer(
            pattern=self,
            context=context,
            clock=clock,
            callback=callback,
            target_bus=target_bus,
            target_node=target_node,
            uuid=uuid,
        )
        player.play(quantization=quantization, at=at, until=until)
        return player

    ### PUBLIC PROPERTIES ###

    @abc.abstractproperty
    def is_infinite(self) -> bool:
        raise NotImplementedError


class BinaryOpPattern(Pattern[T]):
    ### INITIALIZER ###

    def __init__(
        self,
        operator_: Callable,
        expr_one: Union[Pattern[T], T],
        expr_two: Union[Pattern[T], T],
    ) -> None:
        self.operator_ = operator_
        self.expr_one = self._freeze_recursive(expr_one)
        self.expr_two = self._freeze_recursive(expr_two)

    ### PRIVATE METHODS ###

    def _iterate(
        self, state: Optional[Dict[str, UUID]] = None
    ) -> Generator[T, bool, None]:
        iterator_one: Iterator[T] = iter(
            self.expr_one
            if isinstance(self.expr_one, Pattern)
            else SequencePattern([self.expr_one], None)
        )
        iterator_two: Iterator[T] = iter(
            self.expr_two
            if isinstance(self.expr_two, Pattern)
            else SequencePattern([self.expr_two], None)
        )
        for item_one, item_two in zip(iterator_one, iterator_two):
            yield self._apply_recursive(self.operator_, item_one, item_two)

    ### PUBLIC PROPERTIES ###

    @property
    def is_infinite(self) -> bool:
        expr_one_is_infinite = (
            not isinstance(self.expr_one, Pattern) or self.expr_one.is_infinite
        )
        expr_two_is_infinite = (
            not isinstance(self.expr_two, Pattern) or self.expr_two.is_infinite
        )
        return expr_one_is_infinite and expr_two_is_infinite


class UnaryOpPattern(Pattern[T]):
    ### INITIALIZER ###

    def __init__(self, operator_: Callable, expr: Union[Pattern[T], T]) -> None:
        self.operator_ = operator_
        self.expr = expr

    def _iterate(
        self, state: Optional[Dict[str, UUID]] = None
    ) -> Generator[T, bool, None]:
        iterator: Iterator[T] = iter(
            self.expr
            if isinstance(self.expr, Pattern)
            else SequencePattern([self.expr], None)
        )
        for item in iterator:
            yield self._apply_recursive(self.operator_, item)

    ### PUBLIC PROPERTIES ###

    @property
    def is_infinite(self) -> bool:
        return not isinstance(self.expr, Pattern) or self.expr.is_infinite


class SeedPattern(Pattern[T]):
    ### INITIALIZER ###

    def __init__(self, pattern: Pattern[T], seed: int = 0) -> None:
        if not isinstance(pattern, Pattern):
            raise ValueError(f"Must be pattern: {pattern!r}")
        self._pattern = pattern
        self._seed = int(seed)

    ### PRIVATE METHODS ###

    def _iterate(
        self, state: Optional[Dict[str, UUID]] = None
    ) -> Generator[T, bool, None]:
        try:
            identifier = id(inspect.currentframe())
            rng = self._get_seeded_rng(seed=self.seed)
            self._rngs[identifier] = rng
            yield from self._pattern
        finally:
            del self._rngs[identifier]

    ### PUBLIC PROPERTIES ###

    @property
    def is_infinite(self) -> bool:
        return self._pattern.is_infinite

    @property
    def pattern(self) -> Pattern:
        return self._pattern

    @property
    def seed(self) -> int:
        return self._seed


class SequencePattern(Pattern[T]):
    ### INITIALIZER ###

    def __init__(
        self, sequence: Sequence[Union[T, Pattern[T]]], iterations: Optional[int] = 1
    ) -> None:
        if not isinstance(sequence, Sequence):
            raise ValueError(f"Must be sequence: {sequence!r}")
        if iterations is not None:
            iterations = int(iterations)
            if iterations < 1:
                raise ValueError("Iterations must be null or greater than 0")
        self._sequence: Sequence[Union[T, Pattern[T]]] = self._freeze_recursive(
            sequence
        )
        self._iterations = iterations

    ### PRIVATE METHODS ###

    def _iterate(self, state=None):
        should_stop = False
        for _ in self._loop(self._iterations):
            for x in self._sequence:
                if not isinstance(x, Pattern):
                    should_stop = (yield x) or should_stop
                else:
                    iterator = iter(x)
                    try:
                        y = next(iterator)
                        should_stop = (yield y) or should_stop
                        while True:
                            y = iterator.send(should_stop)
                            should_stop = (yield y) or should_stop
                    except StopIteration:
                        pass
                if should_stop:
                    return

    ### PUBLIC PROPERTIES ###

    @property
    def is_infinite(self) -> bool:
        if self._iterations is None:
            return True
        x: Union[T, Pattern[T]]
        for x in self._sequence:
            if isinstance(x, Pattern) and x.is_infinite:
                return True
        return False



# flake8: noqa
import os
import platform

import pytest

from supriya.ugens import Out, SinOsc, SuperColliderSynthDef, SynthDefBuilder


@pytest.fixture
def py_synthdef():
    with SynthDefBuilder() as builder:
        sine_a = SinOsc.ar(frequency=420)
        sine_b = SinOsc.ar(frequency=440)  # noqa
        sine_c = SinOsc.ar(frequency=460)
        sine_d = SinOsc.ar(frequency=sine_c)  # noqa
        Out.ar(bus=0, source=sine_a)
    py_synthdef = builder.build("optimized")
    return py_synthdef


def test_optimization_01_supriya_vs_bytes(py_synthdef):
    # fmt: off
    test_compiled_synthdef = bytes(
        b'SCgf'
        b'\x00\x00\x00\x02'
        b'\x00\x01'
            b'\toptimized'
                b'\x00\x00\x00\x02'
                    b'C\xd2\x00\x00'
                    b'\x00\x00\x00\x00'
                b'\x00\x00\x00\x00'
                b'\x00\x00\x00\x00'
                b'\x00\x00\x00\x02'
                    b'\x06SinOsc'
                        b'\x02'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\xff\xff\xff\xff'
                                b'\x00\x00\x00\x00'
                            b'\xff\xff\xff\xff'
                                b'\x00\x00\x00\x01'
                            b'\x02'
                    b'\x03Out'
                        b'\x02'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x00'
                        b'\x00\x00'
                            b'\xff\xff\xff\xff'
                                b'\x00\x00\x00\x01'
                            b'\x00\x00\x00\x00'
                                b'\x00\x00\x00\x00'
                b'\x00\x00'
    )
    # fmt: on
    py_compiled_synthdef = py_synthdef.compile()
    assert py_compiled_synthdef == test_compiled_synthdef


@pytest.mark.skipif(platform.system() == "Windows", reason="hangs on Windows")
@pytest.mark.skipif(
    platform.system() == "Darwin" and os.environ.get("CI") == "true",
    reason="sclang hangs without QT",
)
def test_optimization_01_supriya_vs_sclang(py_synthdef):
    sc_synthdef = SuperColliderSynthDef(
        "optimized",
        r"""
        var sine_a, sine_b, sine_c, sine_d;
        sine_a = SinOsc.ar(420);
        sine_b = SinOsc.ar(440);
        sine_c = SinOsc.ar(460);
        sine_d = SinOsc.ar(sine_c);
        Out.ar(0, sine_a);
        """,
    )
    sc_compiled_synthdef = bytes(sc_synthdef.compile())
    py_compiled_synthdef = py_synthdef.compile()
    assert py_compiled_synthdef == sc_compiled_synthdef
